from typing import Literal

import numpy as np
import pandas as pd

ModifierType = Literal["Average", "Permutation"]


class ModifierHandler:
    def __init__(self, methods: ModifierType):
        self.methods = methods

    def __call__(self, x, indexes) -> pd.DataFrame:
        results = {}
        for method in self.methods:
            new_x, updated_indexes = self.apply_method(method, x, indexes)
            results[method] = {"x": new_x, "updated_features": updated_indexes}
        return results

    def apply_method(self, method_name: ModifierType, x, indexes) -> pd.DataFrame:
        if method_name == "Average":
            return replace_data_with_average(x, indexes)
        if method_name == "Permutation":
            return replace_data_with_permutation(x, indexes)
        raise NotImplementedError(f"Method {method_name} not implemented")


def replace_data_with_average(x, important_feature_names):
    """
    Replaces the values of columns that are not in `important_feature_names` with the mean value of each column.

    Parameters
    ----------
    x : pandas.DataFrame
        The input DataFrame containing the data to be modified.
    important_feature_names : list
        A list of column names that should not be modified.

    Returns
    -------
    pandas.DataFrame
        A new DataFrame with the modified values.

    Examples
    --------
    >>> import pandas as pd
    >>> data = pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6], "C": [7, 8, 9]})
    >>> important_features = ["A", "C"]
    >>> replace_data_with_average(data, important_features)
       A  B  C
    0  1  6  7
    1  2  6  8
    2  3  6  9
    """
    feature_names = x.columns
    no_important_feature_names = [f for f in feature_names if f not in important_feature_names]

    xdf = x.copy()
    average = x[no_important_feature_names].select_dtypes(include=["number"]).mean()

    xdf[average.index] = average

    return xdf, list(average.index)


def replace_data_with_permutation(x: pd.DataFrame, important_feature_names: list) -> pd.DataFrame:
    """
    Replace non-important column values with random samples chosen from the same column.

    Parameters
    ----------
    x : pd.DataFrame
        The input DataFrame.
    important_feature_names : list
        List of column names that should not be replaced.

    Returns
    -------
    pd.DataFrame
        DataFrame with non-important columns replaced by random samples.

    Example
    -------
    >>> import pandas as pd
    >>> import numpy as np
    >>> df = pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6], "C": [7, 8, 9]})
    >>> important_features = ["A", "C"]
    >>> replace_data_with_permutation(df, important_features)
       A  B  C
    0  1  5  7
    1  2  6  8
    2  3  4  9
    """
    feature_names = x.columns
    no_important_feature_names = [f for f in feature_names if f not in important_feature_names]

    xdf = x.copy()

    for yc in no_important_feature_names:
        unique_values = np.unique(xdf.loc[:, yc])
        xdf.loc[:, yc] = np.random.choice(unique_values, len(xdf))

    return xdf, no_important_feature_names



from holisticai.explainability.metrics._classification import (
    classification_explainability_metrics,
)
from holisticai.explainability.metrics._multiclass import (
    multiclass_explainability_metrics,
)
from holisticai.explainability.metrics._regression import (
    regression_explainability_metrics,
)
from holisticai.explainability.metrics.global_importance import (
    alpha_score,
    position_parity,
    rank_alignment,
    spread_divergence,
    spread_ratio,
    surrogate_accuracy_score,
    surrogate_mean_squared_error,
    xai_ease_score,
)
from holisticai.explainability.metrics.local_importance import (
    data_stability,
    feature_stability,
)

__all__ = [
    "classification_explainability_metrics",
    "multiclass_explainability_metrics",
    "regression_explainability_metrics",
    "alpha_score",
    "position_parity",
    "rank_alignment",
    "spread_ratio",
    "spread_divergence",
    "xai_ease_score",
    "data_stability",
    "feature_stability",
    "surrogate_accuracy_score",
    "surrogate_mean_squared_error",
]



# Imports
import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_approx_equal

# Regression
from holisticai.bias.metrics import (
    avg_score_diff,
    avg_score_ratio,
    correlation_diff,
    disparate_impact_regression,
    jain_index,
    mae_ratio,
    max_statistical_parity,
    no_disparate_impact_level,
    rmse_ratio,
    statistical_parity_auc,
    statistical_parity_regression,
    success_rate_regression,
    zscore_diff,
)

# Formatting
from holisticai.utils import extract_columns
from tests.bias.utils import load_bias_regression_data

                        # Dataset
df_r = load_bias_regression_data()

# Format data into numpy
group_a, group_b, y_pred_r, y_true_r = extract_columns(
    df_r, cols=["group_a", "group_b", "y_pred", "y_true"]
)

def test_success_rate_regression():
    y_pred = [20, 30, 12, 45]
    group_a = [1, 1, 0, 0]
    group_b = [0, 0, 1, 1]
    assert (
        success_rate_regression(group_a, group_b, y_pred, threshold=21)["sr_a"] == 0.5
    )
    assert (
        success_rate_regression(group_a, group_b, y_pred, threshold="median")["sr_b"]
        == 0.5
    )
    with pytest.raises(ValueError) as e_info:
        success_rate_regression(group_a, group_b, y_pred, threshold="typo")


def test_disparate_impact_regression():
    """test disparate_impact_regression"""
    a = disparate_impact_regression(group_a, group_b, y_pred_r, q=0.5)
    b = 1
    assert_approx_equal(a, b)


def test_statistical_parity_regression():
    """test statistical_parity_regression"""
    a = statistical_parity_regression(group_a, group_b, y_pred_r, q=0.5)
    b = 0
    assert_approx_equal(a, b)


def test_disparate_impact_regression90():
    """test disparate_impact_regression90"""
    a = disparate_impact_regression(group_a, group_b, y_pred_r, q=0.9)
    b = 0
    assert_approx_equal(a, b)


def test_disparate_impact_regression80():
    """test disparate_impact_regression80"""
    a = disparate_impact_regression(group_a, group_b, y_pred_r, q=0.8)
    b = 1.5
    assert_approx_equal(a, b)


def test_disparate_impact_regression50():
    """test disparate_impact_regression50"""
    a = disparate_impact_regression(group_a, group_b, y_pred_r, q=0.5)
    b = 1
    assert_approx_equal(a, b)


def test_no_disparate_impact_level():
    """test no_disparate_impact_level"""
    a = no_disparate_impact_level(group_a, group_b, y_pred_r)
    b = 0.7
    assert_approx_equal(a, b)


def test_avg_score_diff():
    """test avg_score_diff"""
    a = avg_score_diff(group_a, group_b, y_pred_r)
    b = 0.1 / 12
    assert_approx_equal(a, b)


def test_avg_score_ratio():
    """test avg_score_ratio"""
    a = avg_score_ratio(group_a, group_b, y_pred_r)
    b = 1.016129032
    assert_approx_equal(a, b)


def test_avg_score_diffQ80():
    """test avg_score_diffQ80"""
    a = avg_score_diff(group_a, group_b, y_pred_r, q=0.8)
    b = -0.1
    assert_approx_equal(a, b)


def test_zscore_diff():
    """test zscore_diff"""
    a = zscore_diff(group_a, group_b, y_pred_r)
    b = 0.025042098285188583
    assert_approx_equal(a, b)


def test_zscore_diffQ80():
    """test zscore_diffQ80"""
    z = np.zeros(40)
    group_a_new = np.concatenate((group_a, z))
    group_b_new = np.concatenate((group_b, z))
    y_pred_new = np.concatenate((y_pred_r, z))
    a = zscore_diff(group_a_new, group_b_new, y_pred_new, q=0.8)
    b = 0.025042098285188583
    assert_approx_equal(a, b)


def test_statistical_parity_auc():
    """test statistical_parity_auc"""
    # test 1
    group_a = np.array([1] * 100 + [0] * 100)
    group_b = np.array([0] * 100 + [1] * 100)
    y_pred = np.array(list(2 * np.linspace(0, 1, 100)) + list(np.linspace(0, 1, 100)))
    a = statistical_parity_auc(group_a, group_b, y_pred)
    b = 0.24753333333333336
    assert_approx_equal(a, b)
    # test 2
    y_pred = np.array(list(np.linspace(0, 1, 100)) + list(2 * np.linspace(0, 1, 100)))
    a = statistical_parity_auc(group_a, group_b, y_pred)
    b = 0.24753333333333336
    assert_approx_equal(a, b)


def test_max_statistical_parity():
    """test max_statistical_parity"""
    # test 1
    group_a = np.array([1] * 100 + [0] * 100)
    group_b = np.array([0] * 100 + [1] * 100)
    y_pred = np.array(list(2 * np.linspace(0, 1, 100)) + list(np.linspace(0, 1, 100)))
    a = max_statistical_parity(group_a, group_b, y_pred)
    b = 0.5
    assert_approx_equal(a, b)


def test_correlation_diff():
    """test correlation_diff"""
    a = correlation_diff(group_a, group_b, y_pred_r, y_true_r)
    b = -0.022
    assert_approx_equal(a, b, 3)


def test_correlation_diffQ80():
    """test correlation_diffQ80"""
    z = np.zeros(40)
    group_a_new = np.concatenate((group_a, z))
    group_b_new = np.concatenate((group_b, z))
    y_pred_new = np.concatenate((y_pred_r, z))
    y_true_new = np.concatenate((y_true_r, z))
    a = correlation_diff(group_a_new, group_b_new, y_pred_new, y_true_new, q=0.8)
    b = -0.022
    assert_approx_equal(a, b, 3)


def test_rmse_ratio():
    """test rmse_ratio"""
    a = rmse_ratio(group_a, group_b, y_pred_r, y_true_r)
    b = 1.3228756555322954
    assert_approx_equal(a, b)


def test_rmse_ratioQ80():
    """test rmse_ratioQ80"""
    z = np.zeros(40)
    group_a_new = np.concatenate((group_a, z))
    group_b_new = np.concatenate((group_b, z))
    y_pred_new = np.concatenate((y_pred_r, z))
    y_true_new = np.concatenate((y_true_r, z))
    a = rmse_ratio(group_a_new, group_b_new, y_pred_new, y_true_new, q=0.8)
    b = 1.3228756555322954
    assert_approx_equal(a, b)


def test_mae_ratio():
    """test mae_ratio"""
    a = mae_ratio(group_a, group_b, y_pred_r, y_true_r)
    b = 1.25
    assert_approx_equal(a, b)


def test_mae_ratioQ80():
    """test mae_ratioQ80"""
    z = np.zeros(40)
    group_a_new = np.concatenate((group_a, z))
    group_b_new = np.concatenate((group_b, z))
    y_pred_new = np.concatenate((y_pred_r, z))
    y_true_new = np.concatenate((y_true_r, z))
    a = mae_ratio(group_a_new, group_b_new, y_pred_new, y_true_new, q=0.8)
    b = 1.25
    assert_approx_equal(a, b)


def test_jain_index():
    """test jain_index"""
    a = jain_index(y_pred_r, y_true_r)
    b = 0.9307692
    assert_approx_equal(a, b)



from __future__ import annotations

from collections import Counter
from typing import Optional, Union

import numpy as np
import pandas as pd
from sklearn.compose import ColumnTransformer
from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import OneHotEncoder
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor


class Anonymize:
    """Anonymize
    Class for performing tailored, model-guided anonymization of training datasets for ML models.

    Parameters
    ----------
    k : int
        The privacy parameter that determines the number of records that will be indistinguishable from each
        other (when looking at the quasi identifiers). Should be at least 2.

    quasi_identifiers :  np.ndarray or list of strings or integers.
        The features that need to be minimized in case of pandas data, and indexes of features
        in case of numpy data.

    quasi_identifer_slices: list of lists of strings or integers.
        If some of the quasi-identifiers represent 1-hot encoded features that need to remain
        consistent after anonymization, provide a list containing the list of column names
        or indexes that represent a single feature.

    categorical_features : list
        The list of categorical features (if supplied, these featurtes will be one-hot encoded
        before using them to train the decision tree model).

    is_regression : bool
        Whether the model is a regression model or not (if False, assumes a classification model).
        Default is False.

    train_only_QI : bool
        The required method to train data set for anonymization. Default is
        to train the tree on all features.


    References
    ----------

    Goldsteen, Abigail, Gilad Ezov, Ron Shmelkin, Micha Moffie, and Ariel Farkash. "Anonymizing machine learning models."
    In International Workshop on Data Privacy Management, pp. 121-136. Cham: Springer International Publishing, 2021.

    """

    def __init__(
        self,
        k: int,
        quasi_identifiers: Union[np.ndarray, list],
        features: Optional[list] = None,
        features_names: Optional[list] = None,
        quasi_identifer_slices: Optional[list] = None,
        categorical_features: Optional[list] = None,
        is_regression: Optional[bool] = False,
        train_only_QI: Optional[bool] = False,
    ):
        if k < 2:
            raise ValueError("k should be a positive integer with a value of 2 or higher")
        if quasi_identifiers is None or len(quasi_identifiers) < 1:
            raise ValueError("The list of quasi-identifiers cannot be empty")

        self.k = k
        self.quasi_identifiers = quasi_identifiers
        self.categorical_features = categorical_features
        self.is_regression = is_regression
        self.train_only_QI = train_only_QI
        self.features_names = features_names
        self.features = features
        self.quasi_identifer_slices = quasi_identifer_slices

    def anonymize(self, X_train, y_train):
        """
        Description
        -----------
        Method for performing model-guided anonymization.

        Parameters
        ----------
        X_train : np.ndarray or pandas DataFrame
            Dataset containing the training data for the model.

        y_train : np.ndarray
            The predictions of the original model on the training data.

        Returns
        -------
        The anonymized training dataset as pandas DataFrame.

        """
        if X_train.shape[1] != 0:
            self.features = list(range(X_train.shape[1]))
        else:
            raise ValueError("No data provided")

        if self.features_names is None:
            # if no names provided, use numbers instead
            self.features_names = self.features

        if not set(self.quasi_identifiers).issubset(set(self.features_names)):
            raise ValueError(
                "Quasi identifiers should bs a subset of the supplied features or indexes in range of "
                "the data columns"
            )
        if self.categorical_features and not set(self.categorical_features).issubset(set(self.features_names)):
            raise ValueError(
                "Categorical features should bs a subset of the supplied features or indexes in range of "
                "the data columns"
            )

        # transform quasi identifiers to indexes
        self.quasi_identifiers = [i for i, v in enumerate(self.features_names) if v in self.quasi_identifiers]
        if self.quasi_identifer_slices:
            temp_list = []
            for qislice in self.quasi_identifer_slices:
                new_slice = [i for i, v in enumerate(self.features_names) if v in qislice]
                temp_list.append(new_slice)
            self.quasi_identifer_slices = temp_list
        if self.categorical_features:
            self.categorical_features = [i for i, v in enumerate(self.features_names) if v in self.categorical_features]

        transformed = self._anonymize(X_train.copy(), y_train)
        df = pd.DataFrame(transformed, columns=self.features_names)
        return df.astype(X_train.dtypes)

    def _anonymize(self, x, y):
        if x.shape[0] != y.shape[0]:
            raise ValueError("x and y should have same number of rows")
        if any(x.select_dtypes(include="category").columns):
            if not self.categorical_features:
                raise ValueError("when supplying an array with non-numeric data, categorical_features must be defined")
            x_prepared = self._modify_categorical_features(x)
        else:
            x_prepared = x
        x_anonymizer_train = x_prepared
        if self.train_only_QI:
            # build DT just on QI features
            x_anonymizer_train = x_prepared[:, self.quasi_identifiers]
        if self.is_regression:
            self._anonymizer = DecisionTreeRegressor(random_state=10, min_samples_split=2, min_samples_leaf=self.k)
        else:
            self._anonymizer = DecisionTreeClassifier(random_state=10, min_samples_split=2, min_samples_leaf=self.k)

        self._anonymizer.fit(x_anonymizer_train, y)
        cells_by_id = self._calculate_cells(x, x_anonymizer_train)
        return self._anonymize_data(x, x_anonymizer_train, cells_by_id)

    def _calculate_cells(self, x, x_anonymizer_train):
        # x is original data, x_anonymizer_train is only QIs + 1-hot encoded
        cells_by_id = {}
        leaves = []
        for node, feature in enumerate(self._anonymizer.tree_.feature):
            if feature == -2:  # leaf node
                leaves.append(node)
                hist = [int(i) for i in self._anonymizer.tree_.value[node][0]]
                # TODO we may change the method for choosing representative for cell  # noqa: TD004
                # label_hist = self.anonymizer.tree_.value[node][0]
                # label = int(self.anonymizer.classes_[np.argmax(label_hist)])
                cell = {"label": 1, "hist": hist, "id": int(node)}
                cells_by_id[cell["id"]] = cell
        self._nodes = leaves
        self._find_representatives(x, x_anonymizer_train, cells_by_id.values())
        return cells_by_id

    def _find_representatives(self, x, x_anonymizer_train, cells):
        # x is original data (always numpy), x_anonymizer_train is only QIs + 1-hot encoded
        node_ids = self._find_sample_nodes(x_anonymizer_train)
        if self.quasi_identifer_slices:
            all_one_hot_features = {feature for encoded in self.quasi_identifer_slices for feature in encoded}
        else:
            all_one_hot_features = set()
        for cell in cells:
            cell["representative"] = {}
            # get all rows in cell
            indexes = [index for index, node_id in enumerate(node_ids) if node_id == cell["id"]]
            # TODO: should we filter only those with majority label? (using hist)
            rows = x.loc[indexes]
            done = set()
            for feature in self.quasi_identifiers:  # self.quasi_identifiers are numerical indexes
                if feature not in done:
                    # deal with 1-hot encoded features
                    if feature in all_one_hot_features:
                        # find features that belong together
                        for encoded in self.quasi_identifer_slices:
                            if feature in encoded:
                                values = rows[:, encoded]
                                unique_rows, counts = np.unique(values, axis=0, return_counts=True)
                                rep = unique_rows[np.argmax(counts)]
                                for i, e in enumerate(encoded):
                                    done.add(e)
                                    cell["representative"][e] = rep[i]
                    else:  # rest of features
                        values = rows.iloc[:, feature]
                        if self.categorical_features and feature in self.categorical_features:
                            # find most common value
                            cell["representative"][feature] = Counter(values).most_common(1)[0][0]
                        else:
                            # find the mean value (per feature)
                            median = np.median(values)
                            min_value = max(values)
                            min_dist = float("inf")
                            for value in values:
                                # euclidean distance between two floating point values
                                dist = abs(value - median)
                                if dist < min_dist:
                                    min_dist = dist
                                    min_value = value
                            cell["representative"][feature] = min_value

    def _find_sample_nodes(self, samples):
        paths = self._anonymizer.decision_path(samples).toarray()
        node_set = set(self._nodes)
        return [next(iter({i for i, v in enumerate(p) if v == 1} & node_set)) for p in paths]

    def _find_sample_cells(self, samples, cells_by_id):
        node_ids = self._find_sample_nodes(samples)
        return [cells_by_id[node_id] for node_id in node_ids]

    def _anonymize_data(self, x, x_anonymizer_train, cells_by_id):
        cells = self._find_sample_cells(x_anonymizer_train, cells_by_id)
        index = 0
        dfnew = pd.DataFrame(columns=x.columns)
        for _, row in x.iterrows():
            cell = cells[index]
            index += 1
            for feature in cell["representative"]:
                row[feature] = cell["representative"][feature]
            dfnew.loc[len(dfnew)] = row
        return dfnew

    def _modify_categorical_features(self, x):
        # prepare data for DT
        used_features = self.features
        if self.train_only_QI:
            used_features = self.quasi_identifiers
        numeric_features = [f for f in self.features if f in used_features and f not in self.categorical_features]
        categorical_features = [f for f in self.categorical_features if f in used_features]
        numeric_transformer = Pipeline(steps=[("imputer", SimpleImputer(strategy="constant", fill_value=0))])

        categorical_transformer = OneHotEncoder(handle_unknown="ignore", sparse_output=False)
        preprocessor = ColumnTransformer(
            transformers=[
                ("num", numeric_transformer, numeric_features),
                ("cat", categorical_transformer, categorical_features),
            ]
        )
        encoded = preprocessor.fit_transform(x)
        return encoded



import logging
import sys
from collections import defaultdict

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import accuracy_score, roc_auc_score
from sklearn.model_selection import train_test_split
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler, StandardScaler

from holisticai.bias.metrics import average_odds_diff, statistical_parity
from holisticai.utils.trade_off_analysers.utils_fairea import (
    do_line_segments_intersect,
    find_segments,
    get_area,
    line,
)

logger = logging.getLogger(__name__)

METRICS = {
    "acc": accuracy_score,
    "auc": roc_auc_score,
    "sp": statistical_parity,
    "aod": average_odds_diff,
}


class Fairea:
    """
    Fairea class for calculating the trade-off between accuracy and fairness of a binary classification model.

    Fairea uses a model behaviour mutation method to create a baseline that can be used to compare quantitatively\
    the fairness-accuracy trade-off for different bias mitigation algorithms and then evaluate their effectiveness\
    in the given scenario. To perform this analysis, the approach consists of three separate steps:

    - A baseline is created by fitting a model without mitigation and then applying a model behaviour\
    mutation to gradually changes the model outputs.
    - Map the given fitted bias mitigation models into five mitigation regions to classify their effectiveness.
    - Assess the effectiveness trade-off by measuring the gap between the mitigators' effectiveness and the baseline.

    Parameters
    ----------
    fair_metric: str
        fairness metric to be used: statistical parity (sp) or average odds difference (aod)
    acc_metric: str
        accuracy metric to be used in the baseline creation and trade-off comparison: accuracy (acc) or roc-auc (auc)
    verbose: bool
        whether to print the progress or not

    Methods
    -------
    create_baseline(x, y, group_a, group_b, test_size, data_splits, repetitions,\
        odds, options, degrees)
        Creates the baseline model for the dataset.
    baseline_metrics()
        Returns the baseline metrics.
    add_model_outcomes(model_name, y_true, y_pred, group_a, group_b)
        Adds the outcomes of a new model to be compared with the baseline model.
    plot_baseline(ax, figsize, title, normalize)
        Plots the baseline model.
    plot_methods(ax, figsize, title, normalize)
        Plots the baseline with the added models.
    trade_off_region_classification()
        Classifies the region of each model.
    compute_trade_off_area()
        Determines the area of each model and saves the best method.
    get_best_model()
        Returns the best model.

    Attributes
    ----------
    verbose: bool
        whether to print the progress or not
    acc_fn: function
        accuracy function to be used
    fair_fn: function
        fairness function to be used
    fair_metric: str
        fairness metric to be used
    methods: dict
        dictionary of the methods added to the class
    normalized_methods: dict
        dictionary of the normalized methods added to the class
    best: str
        best method added to the class
    mitigation_regions: dict
        dictionary of the mitigation regions of the methods added to the class
    odds: dict
        dictionary of odds to be used for mutation
    options: list
        list of options to be used for mutation
    degrees: int
        number of divisions in the range of 0-1 to be used for degrees mutation


    References
    ----------
    ..  [1] Max Hort, Jie M. Zhang, Federica Sarro, and Mark Harman. 2021. Fairea: a model behaviour\
        mutation approach to benchmarking bias mitigation methods. In Proceedings of the 29th ACM\
        Joint Meeting on European Software Engineering Conference and Symposium on the Foundations\
        of Software Engineering (ESEC/FSE 2021)
    """

    def __init__(
        self,
        fair_metric="sp",
        acc_metric="acc",
        verbose=False,  # noqa: FBT002
    ):
        self.verbose = verbose
        self.acc_fn = METRICS[acc_metric]
        self.fair_fn = METRICS[fair_metric]
        self.fair_metric = fair_metric
        self.methods = {}
        self.normalized_methods = {}
        self.best = None
        self.mitigation_regions = {}

    def create_baseline(
        self,
        x,
        y,
        group_a,
        group_b,
        test_size=0.3,
        data_splits=10,
        repetitions=10,
        odds=None,
        options=None,
        degrees=10,
    ):
        """
        Creates the baseline model for the dataset.

        Parameters
        ----------
        x: array, shape=(n_samples, n_features)
            input data
        y: array, shape = (n_samples,)
            target data
        group_a: array, shape = (n_samples,)
            protected attribute for group a
        group_b: array, shape = (n_samples,)
            protected attribute for group b
        test_size: float
            percentage size of the test set (in range 0,1)
        data_splits: int
            number of splits to be used for cross-validation
        repetitions: int
            number of repetitions of mutation to be performed
        odds: dict
            dictionary of odds to be used for mutation
        options: list
            list of options to be used for mutation
        degrees: int
            number of divisions in the range of 0-1 to be used for degrees mutation
        """
        if odds is None:
            self.odds = {"0": [1, 0], "1": [0, 1]}
        if options is None:
            self.options = [0, 1]

        self.degrees = degrees

        # _check_same_shape([group_a, group_b, x, y], names="group_a, group_b, x, y")
        list_of_arr = [group_a, group_b, x, y]
        length = list_of_arr[0].shape[0]

        # check if all arrays have the same length
        if not all(arr.shape[0] == length for arr in list_of_arr):
            msg = "All arrays must have the same length"
            raise ValueError(msg)

        self.baseline_acc, self.baseline_fairness = self.__create_baseline(
            x,
            y,
            group_a,
            group_b,
            test_size=test_size,
            data_splits=data_splits,
            repetitions=repetitions,
        )
        self.acc_scaler = MinMaxScaler()
        self.fair_scaler = MinMaxScaler()
        self.acc_norm = self.acc_scaler.fit_transform(self.baseline_acc).reshape(-1)
        self.fairness_norm = self.fair_scaler.fit_transform(self.baseline_fairness).reshape(-1)

    def baseline_metrics(self):
        """
        Returns the baseline metrics.

        Returns
        -------
        baseline_acc: array, shape = (n_samples,)
            baseline accuracy
        baseline_fairness: array, shape = (n_samples,)
            baseline fairness
        """
        return self.baseline_acc, self.baseline_fairness

    def __mutate_preds(self, preds, to_mutate, ids, odds, options):
        """
        Mutates the predictions of the model.

        Parameters
        ----------
        preds: array, shape = (n_samples,)
            predictions of the model
        to_mutate: int
            number of labels to mutate
        ids: list
            ids of the predictions to be mutated
        odds: list
            odds to be used for mutation
        options: list
            list of options to be used for mutation

        Returns
        -------
        changed: array, shape = (n_samples,)
            mutated predictions
        """
        rand = np.random.choice(options, to_mutate, p=odds)
        # Select prediction ids that are being mutated
        to_change = np.random.choice(ids, size=to_mutate, replace=False)
        changed = np.copy(preds)
        for t, r in zip(to_change, rand):
            changed[t] = r
        return changed

    def __create_baseline(
        self,
        x,
        y,
        group_a,
        group_b,
        test_size,
        data_splits,
        repetitions,
    ):
        """
        Creates the baseline model for the dataset by applying the mutations.

        Parameters
        ----------
        x: array, shape=(n_samples, n_features)
            input data
        y: array, shape = (n_samples,)
            target data
        group_a: array, shape = (n_samples,)
            protected attribute for group a
        group_b: array, shape = (n_samples,)
            protected attribute for group b
        test_size: float
            percentage size of the test set (in range 0,1)
        data_splits: int
            number of splits to be used for cross-validation
        repetitions: int
                    number of repetitions of mutation to be performed

        Returns
        -------
        acc_base: array, shape = (n_samples,)
            baseline accuracy
        """
        n_samples = int(x.shape[0] * test_size)

        ids = list(range(n_samples))

        results = defaultdict(lambda: defaultdict(list))

        for s in range(data_splits):
            if self.verbose:
                logger.info("Current datasplit: {s}")

            np.random.seed(s)
            (
                x_train,
                x_test,
                y_train,
                y_test,
                _,
                group_a_ts,
                _,
                group_b_ts,
            ) = train_test_split(x, y, group_a, group_b, test_size=test_size)
            # Define the pipeline
            pipe = Pipeline([("scaler", StandardScaler()), ("clf", LogisticRegression())])

            # Fit the pipeline and make predictions
            pipe.fit(x_train, y_train)
            pred = pipe.predict(x_test).reshape(-1, 1)

            degrees_ = np.linspace(0, 1, self.degrees + 1)

            # Mutate labels for each degree
            for degree in degrees_:
                # total number of labels to mutate
                to_mutate = int(n_samples * degree)

                for name, o in self.odds.items():
                    # Store each mutation attempt
                    hist = []
                    for _ in range(repetitions):
                        # Generate mutated labels
                        mutated = self.__mutate_preds(pred, to_mutate, ids, o, self.options)
                        # Determine accuracy and fairness of mutated model
                        acc = self.acc_fn(y_test, mutated)
                        if self.fair_metric == "aod":
                            fair = abs(self.fair_fn(group_a_ts, group_b_ts, mutated, y_test))
                        else:
                            fair = abs(self.fair_fn(group_a_ts, group_b_ts, mutated))
                        hist.append([acc, fair])
                    results[name][degree] += hist

        acc_base = np.array([np.mean([row[0] for row in results["0"][degree]]) for degree in degrees_])
        fair_base = np.array([np.mean([row[1] for row in results["0"][degree]]) for degree in degrees_])
        return acc_base.reshape(-1, 1), fair_base.reshape(-1, 1)

    def add_model_outcomes(self, model_name, y_true, y_pred, group_a, group_b):
        """
        Adds the outcomes of a new model to be compared with the baseline model.

        Parameters
        ----------
        model_name: str
                name of the model to be added
        y_true: array, shape = (n_samples,)
                target data
        y_pred: array, shape = (n_samples,)
                predictions of the model
        group_a: array, shape = (n_samples,)
                protected attribute for group a
        group_b: array, shape = (n_samples,)
                protected attribute for group b
        """
        list_of_arr = [group_a, group_b, y_pred, y_true]
        length = list_of_arr[0].shape[0]

        # check if all arrays have the same length
        if not all(arr.shape[0] == length for arr in list_of_arr):
            msg = "All arrays must have the same length"
            raise ValueError(msg)

        acc = self.acc_fn(y_true, y_pred)
        if self.fair_metric == "aod":
            fair = abs(self.fair_fn(group_a, group_b, y_pred, y_true))
        else:
            fair = abs(self.fair_fn(group_a, group_b, y_pred))
        self.methods[model_name] = (fair, acc)
        self.normalized_methods[model_name] = (
            self.fair_scaler.transform(np.array([fair]).reshape(1, -1)).reshape(-1)[0],
            self.acc_scaler.transform(np.array([acc]).reshape(1, -1)).reshape(-1)[0],
        )

    def plot_baseline(self, ax=None, figsize=None, title=None, normalize=False):  # noqa: FBT002
        """
        Plots the baseline model.

        Parameters
        ----------
        cmap: string
            color map to be used for plotting
        ax: matplotlib axis
            axis to be used for plotting
        figsize: tuple
            size of the plot
        title: string
            title of the plot
        normalize: boolean
            whether to normalize the data or not
        """
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        ax.set_title(title)
        ax.set_xlabel("Fairness")
        ax.set_ylabel("Accuracy")
        if normalize:
            ax.plot(
                self.fairness_norm,
                self.acc_norm,
                marker="|",
                linestyle="solid",
                linewidth=3,
                markersize=16,
                label="Baseline",
            )
        else:
            ax.plot(
                self.baseline_fairness,
                self.baseline_acc,
                marker="|",
                linestyle="solid",
                linewidth=3,
                markersize=16,
                label="Baseline",
            )
        ax.set_xlim(0)
        ax.legend(loc="best")
        return ax

    def plot_methods(self, ax=None, figsize=None, title=None, normalize=False):  # noqa: FBT002
        """
        Plots the baseline with the added models.


        Parameters
        ----------
        cmap: string
            color map to be used for plotting
        ax: matplotlib axis
            axis to be used for plotting
        figsize: tuple
            size of the plot
        title: string
            title of the plot
        normalize: boolean
            whether to normalize the data or not
        """
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        ax.set_title(title)
        ax.set_xlabel("Fairness")
        ax.set_ylabel("Accuracy")
        if normalize:
            ax.plot(
                self.fairness_norm,
                self.acc_norm,
                marker="|",
                linestyle="solid",
                linewidth=3,
                markersize=16,
                label="Baseline",
            )
        else:
            ax.plot(
                self.baseline_fairness,
                self.baseline_acc,
                marker="|",
                linestyle="solid",
                linewidth=3,
                markersize=16,
                label="Baseline",
            )
        if normalize:
            for name, (fair, acc) in self.normalized_methods.items():
                ax.scatter(fair, acc, color="red")
                ax.annotate(name, (fair, acc))
        else:
            for name, (fair, acc) in self.methods.items():
                ax.scatter(fair, acc, color="red")
                ax.annotate(name, (fair, acc))
        ax.set_xlim(0)
        ax.legend(loc="best")
        return ax


    def determine_region(self, p, fairness_norm, acc_norm):
        """
        Determines the region of the model.

        Parameters
        ----------
        p: tuple
            point of the model
        fairness_norm: list
            normalized fairness of the model
        acc_norm: list
            normalized accuracy of the model

        Returns
        -------
        region: str
            region of the model
        """
        fair_segment, acc_segment = find_segments(p, fairness_norm, acc_norm)
        # Extend bias mitigation point towards four directions (left,right,up,down)
        line_down = line([p[0], p[1]], [p[0], 0])
        line_right = line([p[0], p[1]], [2, p[1]])
        # Determine bias mitigation region based on intersection with baseline
        if p[0] > 1 and p[1] > 1:
            return "inverted"
        if p[0] > 1 and p[1] < 1:
            return "lose-lose"
        if p[0] < 1 and p[1] > 1:
            return "win-win"
        if p[0] >= 0 and p[0] <= 1 and p[1] >= 0 and p[1] <= 1:
            if do_line_segments_intersect(line_down, fair_segment) and do_line_segments_intersect(
                line_right, acc_segment
            ):
                return "good trade-off"
            return "bad trade-off"
        if p[0] < 0:
            return "lose-lose"
        return "inverted"

    def trade_off_region_classification(self, plot=False):  # noqa: FBT002
        """
        Classifies the region of each model.

        Returns
        -------
        df : DataFrame
            dataframe that contains the region of each model of the class
        """
        for k, (fair, acc) in self.normalized_methods.items():
            # define a point for each bias mitigation method
            p = (fair, acc)
            self.mitigation_regions[k] = self.determine_region(p, self.fairness_norm, self.acc_norm)
        # create a dataframe for the region of each method
        df = pd.DataFrame.from_dict(self.mitigation_regions, orient="index", columns=["Region"])
        df.style.set_properties(**{"font-weight": "bold"})
        if plot:
            ax = self.plot_regions(list(zip(self.fairness_norm, self.acc_norm)))
            for name, (fair, acc) in self.normalized_methods.items():
                ax.scatter(fair, acc, color="red")
                ax.annotate(name, (fair, acc))
            return ax
        return df

    def plot_regions(self, polyline):
        # Sort the polyline points by x-coordinate
        polyline.sort(key=lambda p: p[0])

        # Create a figure and a set of subplots
        fig, ax = plt.subplots()

        # Plot the polyline
        polyline_x, polyline_y = zip(*polyline)
        ax.plot(polyline_x, polyline_y, color='black', label='Baseline')

        # Fill the regions
        ax.fill_between(polyline_x, polyline_y, polyline_y[-1], where=[y <= polyline_y[-1] for y in polyline_y],
                        color='green', alpha=0.2, hatch='//', label='Good trade-off')
        ax.fill_between(polyline_x, polyline_y, 0, where=[y >= 0 for y in polyline_y], color='red', alpha=0.2,
                        hatch='\\\\', label='Bad trade-off')
        ax.fill_between(polyline_x, polyline_y[-1], polyline_y[-1]+1, color='blue', alpha=0.2, hatch='--',
                        label='Win-win')
        ax.fill_between(np.arange(polyline_x[-1], polyline_x[-1] + 1, 0.01), 0, polyline_y[-1], color='orange',
                        alpha=0.2, hatch='++', label='Lose-lose')
        ax.fill_between(np.arange(polyline_x[-1], polyline_x[-1]+1, 0.01), polyline_y[-1], polyline_y[-1]+1,
                        color='purple', alpha=0.2, hatch='**', label='Inverted')


        # Set the x and y limits
        ax.set_xlim(0, polyline_x[-1]+1)
        ax.set_ylim(0, polyline_y[-1]+1)

        # x-axis label
        ax.set_xlabel('Normalised Fairness')

        # y-axis label
        ax.set_ylabel('Normalised Accuracy')

        # Place the legend outside the plot
        ax.legend(loc='center left', bbox_to_anchor=(1, 0.5))
        return ax

    def compute_trade_off_area(self):
        """
        Determines the area of each model and saves the best method.

        Returns
        -------
        df : DataFrame
            dataframe that contains the area of each model of the class
        """
        good = {k for k, v in self.mitigation_regions.items() if v == "good trade-off"}
        if not good:
            # sys.stdout.write("There are not good methods, consider using a different fairness metric")
            methods = [v for _, v in self.mitigation_regions.items() if v == "win-win"]
            if all(method == "win-win" for method in methods):
                sys.stdout.write("You can choose one of the win-win methods")
            else:
                sys.stdout.write("There are not good methods, consider using a different fairness metric")
            return
        area_methods = {}
        for item in good:
            mitigator_point = self.normalized_methods[item]
            area = get_area(mitigator_point, self.fairness_norm, self.acc_norm)
            area_methods[item] = area
        # create a dataframe for the area of each method
        df = pd.DataFrame.from_dict(area_methods, orient="index", columns=["Area"])
        # sort the dataframe based on the Area
        df = df.sort_values(by="Area", ascending=False)
        df.style.set_properties(**{"font-weight": "bold"})
        # save method with the largest area in the best variable
        self.best = df.index[0]
        return df

    def get_best_model(self):
        """
        Returns the best model.

        Returns
        -------
        best: str
            best method added to the class
        """
        if self.best is None:
            methods = [v for _, v in self.mitigation_regions.items() if v == "win-win"]
            if all(method == "win-win" for method in methods):
                logger.warning("You can choose one of the win-win methods")
            else:
                logger.warning("There are not good methods, consider using a different fairness metric")
        return self.best



from holisticai.utils import LocalImportances
from holisticai.explainability.metrics import data_stability
import pandas as pd
import numpy as np

def test_local_feature_importance():
    data = pd.DataFrame.from_dict({
        "feature1": [1, 2, 3, 4],
        "feature2": [5, 6, 7, 8],
        "feature3": [9, 10, 11, 12],
        "feature4": [13, 14, 15, 16]
    })
    local_importance = LocalImportances(data=data, cond= pd.Series([1, 2, 1, 2]))
    conditional_local_importance = local_importance.conditional()
    assert len(conditional_local_importance) == 2

def test_local_feature_importance_to_global():
    data = pd.DataFrame.from_dict({
        "feature1": [1, 2, 3, 4],
        "feature2": [5, 6, 7, 8],
        "feature3": [9, 10, 11, 12],
        "feature4": [13, 14, 15, 16]
    })
    local_importance = LocalImportances(data=data, cond= pd.Series([1, 2, 1, 2]))
    assert len(local_importance.to_global().values)==4


def test_data_stability():
    importances = pd.DataFrame({
          "feature_1": [0.10, 0.20, 0.30],
          "feature_2": [0.10, 0.25, 0.35],
          "feature_3": [0.15, 0.20, 0.30]})
    local_importances = LocalImportances(importances)
    stability_score = data_stability(local_importances)
    assert np.isclose(stability_score, 0.9900968679601143)

def test_partial_dependence():
    from holisticai.utils import PartialDependence, Importances
    from holisticai.explainability.metrics.global_importance import xai_ease_score

    partial_dependence = [[
         {'average': [[0.1, 0.2, 0.3, 0.1, 0.2, 0.3, 0.1, 0.2, 0.3]], 'grid_values': [[1,2,3,4,5,6,7,8,9]]},
         {'average': [[0.4, 0.5, 0.6, 0.6, 0.5, 0.4, 0.4, 0.5, 0.6]], 'grid_values': [[1,2,3,4,5,6,7,8,9]]},
    ]]
    partial_dependence = PartialDependence(values=partial_dependence)
    feature_importance = Importances(values=np.array([0.5, 0.5]), feature_names=['feature1', 'feature2'])
    score = xai_ease_score(partial_dependence, feature_importance)
    assert score==0.5


import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

from holisticai.datasets import Dataset

# utils
from holisticai.utils._validation import (
    _check_columns,
    _check_numerical_dataframe,
)


def correlation_matrix_plot(
    dataset: Dataset,
    features,
    target_feature,
    fixed_features=None,
    n_features=10,
    cmap="YlGnBu",
    ax=None,
    size=None,
    title=None,
):
    """Plot the correlation matrix of a given dataframe with respect to
    a given target and a certain number of features.

    Obs. The dataframe must contain only numerical features.

    Parameters
    ----------
    df : (DataFrame)
        Pandas dataframe of the data
    target_feature : (str)
        Column name of the target feature

    n_features (optional) : (int)
        Number of features to plot with the closest correlation to the target
    cmap (optional) : (str)
        Color map to use
    ax (optional) : matplotlib axes
        Pre-existing axes for the plot
    size (optional) : (int, int)
        Size of the figure
    title (optional) : (str)
        Title of the figure

    Returns
    -------
    matplotlib ax
    """
    """Prints the correlation matrix """

    if target_feature not in features:
        features.append(target_feature)

    if fixed_features is None:
        fixed_features = []

    for f in fixed_features:
        if f not in features:
            features.append(f)

    df = _check_numerical_dataframe(dataset[features])
    _check_columns(df, target_feature)

    sns.set_theme(font_scale=1.25)
    if ax is None:
        fig, ax = plt.subplots(figsize=size)

    corrmat = df.corr()
    cols = list(corrmat.nlargest(n_features, target_feature)[target_feature].index)
    for f in fixed_features:
        if f in cols:
            cols.remove(f)
    cols += fixed_features
    cm = np.corrcoef(df[cols].values.T)
    sns.heatmap(
        cm,
        cbar=False,
        annot=True,
        fmt=".2f",
        annot_kws={"size": 10},
        yticklabels=cols,
        xticklabels=cols,
        cmap=cmap,
        ax=ax,
    )
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor")

    if title is not None:
        ax.set_title(title)
    else:
        ax.set_title("Correlation matrix")
    return ax



from holisticai.datasets import load_dataset
from sklearn.linear_model import LinearRegression
from holisticai.explainability.metrics import regression_explainability_metrics
from holisticai.explainability.metrics.global_importance import alpha_score, rank_alignment, position_parity, xai_ease_score
import numpy as np
import pytest

ATOL = 5e-2
@pytest.fixture
def input_data():
    dataset = load_dataset('us_crime').sample(n=1000, random_state=42)
    dataset = dataset.train_test_split(test_size=0.2, random_state=42)
    train = dataset['test']
    test = dataset['test']

    model = LinearRegression()
    model.fit(train['X'], train['y'])
    return model, test

def get_regression_features(model, test):
    from holisticai.utils import RegressionProxy
    from holisticai.utils.feature_importances import compute_permutation_feature_importance
    from holisticai.utils.inspection import compute_partial_dependence
    
    proxy = RegressionProxy(predict=model.predict)
    importances  = compute_permutation_feature_importance(X=test['X'], y=test['y'], proxy=proxy)
    ranked_importances = importances.top_alpha(0.8)
    partial_dependencies = compute_partial_dependence(test['X'], features=ranked_importances.feature_names, proxy=proxy)
    conditional_importances  = compute_permutation_feature_importance(X=test['X'], y=test['y'], proxy=proxy, conditional=True)
    return proxy, importances, ranked_importances, conditional_importances, partial_dependencies

def test_xai_regression_metrics(input_data):
    model, test = input_data
    proxy, importances, ranked_importances, conditional_importances, partial_dependencies = get_regression_features(model, test)

    metrics = regression_explainability_metrics(importances, partial_dependencies, conditional_importances, X=test['X'], y_pred=proxy.predict(test['X']))
    assert np.isclose(metrics.loc['Rank Alignment'].value, 0.7317350088183421, atol=ATOL)
    assert np.isclose(metrics.loc['Position Parity'].value, 0.18504188712522046, atol=ATOL)
    assert np.isclose(metrics.loc['XAI Ease Score'].value, 1.0, atol=ATOL)
    assert np.isclose(metrics.loc['Alpha Importance Score'].value, 0.0891089108910891, atol=ATOL)

def test_xai_classification_metrics_separated(input_data):
    model, test = input_data

    proxy, importances, ranked_importances, conditional_importances, partial_dependencies = get_regression_features(model, test)
       
    value = rank_alignment(conditional_importances, ranked_importances)
    assert np.isclose(value, 0.7317350088183421, atol=ATOL)

    value = position_parity(conditional_importances, ranked_importances)
    assert np.isclose(value, 0.18504188712522046, atol=ATOL)
    
    value = xai_ease_score(partial_dependencies, ranked_importances)
    assert np.isclose(value, 1.0, atol=ATOL)

    value = alpha_score(importances)
    assert np.isclose(value, 0.0891089108910891, atol=ATOL)




# imports
from holisticai.bias.mitigation.inprocessing.adversarial_debiasing.transformer import AdversarialDebiasing
from holisticai.bias.mitigation.inprocessing.exponentiated_gradient.transformer import ExponentiatedGradientReduction
from holisticai.bias.mitigation.inprocessing.fair_k_center_clustering.transformer import FairKCenterClustering
from holisticai.bias.mitigation.inprocessing.fair_k_mediam_clustering.transformer import FairKMedianClustering
from holisticai.bias.mitigation.inprocessing.fairlet_clustering.transformer import FairletClustering
from holisticai.bias.mitigation.inprocessing.grid_search.transformer import GridSearchReduction
from holisticai.bias.mitigation.inprocessing.matrix_factorization.blind_spot_aware import BlindSpotAwareMF
from holisticai.bias.mitigation.inprocessing.matrix_factorization.debiasing_learning.transformer import (
    DebiasingLearningMF,
)
from holisticai.bias.mitigation.inprocessing.matrix_factorization.popularity_propensity import PopularityPropensityMF
from holisticai.bias.mitigation.inprocessing.meta_fair_classifier.transformer import MetaFairClassifier
from holisticai.bias.mitigation.inprocessing.prejudice_remover.transformer import PrejudiceRemover
from holisticai.bias.mitigation.inprocessing.two_sided_fairness.transformer import FairRec
from holisticai.bias.mitigation.inprocessing.variational_fair_clustering.transformer import VariationalFairClustering

__all__ = [
    "ExponentiatedGradientReduction",
    "GridSearchReduction",
    "PrejudiceRemover",
    "MetaFairClassifier",
    "VariationalFairClustering",
    "FairKCenterClustering",
    "FairKMedianClustering",
    "FairletClustering",
    "BlindSpotAwareMF",
    "DebiasingLearningMF",
    "PopularityPropensityMF",
    "FairRec",
    "AdversarialDebiasing",
]
import importlib.util

cvxpy_spec = importlib.util.find_spec("cvxpy")
if cvxpy_spec is not None:
    from holisticai.bias.mitigation.inprocessing.fair_scoring_classifier.transformer import FairScoreClassifier

    __all__ += ["FairScoreClassifier"]



from __future__ import annotations

from typing import Union

import numpy as np
import pandas as pd
from sklearn.metrics import accuracy_score, mean_squared_error


class DataMinimizationAccuracyRatio:
    reference: float = np.inf
    name: str = "Data Minimization Accuracy Ratio"

    def __call__(self, y_true, y_pred, y_pred_dm, return_results=False):
        metrics_results = pd.DataFrame(
            [
                {
                    "Selection Type": yp["selector_type"],
                    "Modifier Type": yp["modifier_type"],
                    "N_feats": yp["n_feats"],
                    "Feats": yp["feats"],
                    "Score": relative_performance(accuracy_score, y_pred, yp["predictions"], y_true=y_true),
                    "Accuracy": accuracy_score(y_true, yp["predictions"]),
                }
                for yp in y_pred_dm
            ]
        )
        index = metrics_results["Score"].argmin()
        if return_results:
            pred_row = pd.DataFrame(
                [
                    {
                        "Selection Type": "Base",
                        "Modifier Type": "Base",
                        "N_feats": 0,
                        "Feats": [],
                        "Score": 1,
                        "Accuracy": accuracy_score(y_true, y_pred),
                    }
                ]
            )
            metrics_results = pd.concat([metrics_results, pred_row], ignore_index=True)
            return metrics_results, metrics_results["Score"].loc[index]
        return float(metrics_results["Score"].loc[index])


class DataMinimizationMSERatio:
    reference: float = 0
    name: str = "Data Minimization MSE Ratio"

    def __call__(self, y_true, y_pred, y_pred_dm, return_results=False):
        metrics_results = pd.DataFrame(
            [
                {
                    "Selection Type": yp["selector_type"],
                    "Modifier Type": yp["modifier_type"],
                    "N_feats": yp["n_feats"],
                    "Feats": yp["feats"],
                    "Score": relative_performance(mean_squared_error, y_pred, yp["predictions"], y_true=y_true),
                    "MSE": mean_squared_error(y_true, yp["predictions"]),
                }
                for yp in y_pred_dm
            ]
        )
        index = metrics_results["Score"].argmin()
        if return_results:
            pred_row = pd.DataFrame(
                [
                    {
                        "Selection Type": "Base",
                        "Modifier Type": "Base",
                        "N_feats": 0,
                        "Feats": [],
                        "Score": 1,
                        "MSE": mean_squared_error(y_true, y_pred),
                    }
                ]
            )
            metrics_results = pd.concat([metrics_results, pred_row], ignore_index=True)
            return metrics_results, metrics_results["Score"].loc[index]
        return float(metrics_results["Score"].loc[index])


def get_learning_task(y_true: pd.Series):
    if y_true.dtype.kind in ["i", "u", "O"]:
        return "classification"
    if y_true.dtype.kind in ["f"]:
        return "regression"
    raise ValueError(f"Unknown learning task. dtype: {y_true.dtype.kind}")


def data_minimization_score(
    y_true: pd.Series,
    y_pred: pd.Series,
    y_pred_dm: dict[str, pd.Series],
    return_results=False,
    learning_task: Union[str, None] = None,
):
    """
    Calculate the accuracy ratio for data minimization. The accuracy ratio is the ratio of the accuracy of the data minimization model to the accuracy of the original model.

    Parameters
    ----------
    y_true: pd.Series
        The true labels.
    y_pred: pd.Series
        The predicted labels.
    y_pred_dm: dict[str, pd.Series]
        The predicted labels for each data minimization technique.
    return_results: bool
        Whether to return the results or not. Default is False.
    learning_task: str (Optional)
        The learning task. Can be either "classification" or "regression". If None, it will be inferred from the data.
    Returns
    -------
        float: The accuracy ratio for data minimization.
        pd.DataFrame: The results of the data minimization if return_results is True.
    """
    if learning_task is None:
        learning_task = get_learning_task(y_true)

    if learning_task == "classification":
        dm = DataMinimizationAccuracyRatio()
    if learning_task == "regression":
        dm = DataMinimizationMSERatio()
    return dm(y_true, y_pred, y_pred_dm, return_results)


def relative_performance(metric_fn, y_pred, y_pred_dm, y_true):
    """
    Parameters
    ----------
    metric_fn: function
        metric function used to compare.

    y_true: array-like
        vector-target

    y_pred_dm: array-like
        predicted vector fitted with data minimization

    y_pred: array-like
        predicted vector fitted with all features

    Return
    ------
        relative performance metric
    """
    y = np.array(y_true).flatten()
    y_pred = np.array(y_pred).flatten()
    y_pred_dm = np.array(y_pred_dm).flatten()
    return metric_fn(y, y_pred) / metric_fn(y, y_pred_dm)


def relative_clustering_performance(metric_fn, y_pred, y_pred_dm, x):
    """
    Parameters
    ----------
    metric_fn: function
        metric function used to compare.

    y_true: array-like
        vector-target

    y_pred_dm: array-like
        predicted vector fitted with data minimization

    X: array-like
        input matrix

    Return
    ------
        relative performance metric
    """
    y_pred = np.array(y_pred).flatten()
    y_pred_dm = np.array(y_pred_dm).flatten()
    try:
        dn = metric_fn(x, y_pred) / metric_fn(x, y_pred_dm)
    except:  # noqa: E722
        dn = np.nan

    return dn



import warnings
import numpy as np
from sklearn.linear_model import LogisticRegression, LinearRegression

from holisticai.pipeline import Pipeline
from holisticai.bias.metrics import classification_bias_metrics, regression_bias_metrics, multiclass_bias_metrics, clustering_bias_metrics, recommender_bias_metrics
from tests.bias.mitigation.utils import (
    check_results,
    categorical_dataset,
    regression_dataset,
    multiclass_dataset,
    clustering_dataset,
    recommender_dataset
)
import pytest

warnings.filterwarnings("ignore")

seed = 42

from holisticai.bias.mitigation import MITIGATOR_NAME
from holisticai.bias.mitigation import ExponentiatedGradientReduction
from holisticai.bias.mitigation import GridSearchReduction
from holisticai.bias.mitigation import MetaFairClassifier
from holisticai.bias.mitigation import PrejudiceRemover
from holisticai.bias.mitigation import FairKCenterClustering
from holisticai.bias.mitigation import FairKMedianClustering
from holisticai.bias.mitigation import FairletClustering
from holisticai.bias.mitigation import VariationalFairClustering
from holisticai.bias.mitigation import PopularityPropensityMF
from holisticai.bias.mitigation import BlindSpotAwareMF
from holisticai.bias.mitigation import FairRec
from holisticai.bias.mitigation import DebiasingLearningMF
from holisticai.bias.mitigation import AdversarialDebiasing

def get_inprocessor(mitigator_name : MITIGATOR_NAME = "CalibratedEqualizedOdds", parameters: dict = {}):
    if mitigator_name == "ExponentiatedGradientReduction":
        return ExponentiatedGradientReduction(**parameters)
    if mitigator_name == "AdversarialDebiasing":
        return AdversarialDebiasing(**parameters)
    elif mitigator_name == "GridSearchReduction":
        return GridSearchReduction(**parameters)
    elif mitigator_name == "MetaFairClassifier":
        return MetaFairClassifier(**parameters)
    elif mitigator_name == "PrejudiceRemover":
        return PrejudiceRemover(**parameters)
    elif mitigator_name == "FairScoreClassifier":
        from holisticai.bias.mitigation import FairScoreClassifier
        return FairScoreClassifier(**parameters)
    elif mitigator_name == "FairKCenterClustering":
        return FairKCenterClustering(**parameters)
    elif mitigator_name == "FairKMedianClustering":
        return FairKMedianClustering(**parameters)
    elif mitigator_name == "FairletClustering":
        return FairletClustering(**parameters)
    elif mitigator_name == "VariationalFairClustering":
        return VariationalFairClustering(**parameters)
    elif mitigator_name == "PopularityPropensityMF":
        return PopularityPropensityMF(**parameters)
    elif mitigator_name == "BlindSpotAwareMF":
        return BlindSpotAwareMF(**parameters)
    elif mitigator_name == "FairRec":
        return FairRec(**parameters)
    elif mitigator_name == "DebiasingLearningMF":
        return DebiasingLearningMF(**parameters)
    else:
        raise NotImplementedError

@pytest.mark.skip(reason="remove CBC dependency to enable this test")
@pytest.mark.parametrize("mitigator_name, mitigator_params, model_params", [
    ("FairScoreClassifier", {'objectives':'ab', 'constraints':{}}, {"random_state":42}),
])
def test_multiclass_inprocessor(mitigator_name, mitigator_params, model_params, multiclass_dataset):
    metrics1 = run_inprocessing_categorical(multiclass_dataset, multiclass_bias_metrics, LogisticRegression, mitigator_name, model_params, mitigator_params, is_multiclass=True)
    metrics2 = run_inprocessing_categorical_pipeline(multiclass_dataset, multiclass_bias_metrics, LogisticRegression, mitigator_name, model_params, mitigator_params, is_multiclass=True)
    check_results(metrics1, metrics2)


@pytest.mark.parametrize("mitigator_name, mitigator_params, model_params", [
    ("ExponentiatedGradientReduction", {"seed":1}, {"random_state":42}),
    ("GridSearchReduction", {},  {"random_state":42}),
    #("MetaFairClassifier", {"seed":1}, {"random_state":42}),
    ("PrejudiceRemover", {}, {"random_state":42}),
])
def test_categorical_inprocessor(mitigator_name, mitigator_params, model_params, categorical_dataset):
    metrics1 = run_inprocessing_categorical(categorical_dataset, classification_bias_metrics, LogisticRegression, mitigator_name, model_params, mitigator_params)
    metrics2 = run_inprocessing_categorical_pipeline(categorical_dataset, classification_bias_metrics, LogisticRegression, mitigator_name, model_params, mitigator_params)
    check_results(metrics1, metrics2)


def test_adversarial_debiasing_inprocessor(categorical_dataset):
    train = categorical_dataset['train']
    test = categorical_dataset['test']
    features_dim = train['X'].shape[1]
    mitigator_params = {"features_dim":features_dim , "keep_prob":0.1, "verbose":1, "learning_rate":0.01,"adversary_loss_weight":3, "print_interval":100, "batch_size":1024, "use_debias": True, "epochs":10, "seed":1}
    mitigator_name = "AdversarialDebiasing"

    x_train = train['X'].astype(np.float64)
    x_test = test['X'].astype(np.float64)
    inp = get_inprocessor(mitigator_name, mitigator_params)
    inp.transform_estimator()
    inp.fit(X=x_train, y=train['y'], group_a=train['group_a'], group_b=train['group_b'])
    y_pred = inp.predict(X=x_test, group_a=test['group_a'], group_b=test['group_b'])
    metrics1 = classification_bias_metrics(test['group_a'], test['group_b'], y_pred, test['y'])


    inp = get_inprocessor(mitigator_name, mitigator_params)
    inp.transform_estimator()

    pipeline = Pipeline(steps=[("bm_inprocessing", inp),])
    pipeline.fit(X=x_train, y=train['y'], bm__group_a=train['group_a'], bm__group_b=train['group_b'])
    y_pred = pipeline.predict(X=x_test, bm__group_a=test['group_a'], bm__group_b=test['group_b'])
    metrics2 = classification_bias_metrics(test['group_a'], test['group_b'], y_pred, test['y'])

    check_results(metrics1, metrics2)

@pytest.mark.parametrize("mitigator_name, mitigator_params, model_params", [
    ("ExponentiatedGradientReduction", {'seed':1, "constraints":"BoundedGroupLoss", "loss":"Absolute", "min_val":-0.1, "max_val":1.3, "upper_bound":0.001}, {}),
    ("GridSearchReduction", {}, {}),
])
def test_regression_inprocessor(mitigator_name, mitigator_params, model_params, regression_dataset):
    metrics1 = run_inprocessing_categorical(regression_dataset, regression_bias_metrics, LinearRegression, mitigator_name, model_params, mitigator_params)
    metrics2 = run_inprocessing_categorical_pipeline(regression_dataset, regression_bias_metrics, LinearRegression, mitigator_name, model_params, mitigator_params)
    check_results(metrics1, metrics2)

@pytest.mark.parametrize("mitigator_name, mitigator_params", [
    ("FairKCenterClustering", {"seed":42, "req_nr_per_group":(1,1)}),
    ("FairKMedianClustering", {"seed":42, "n_clusters":2, "strategy":"GA"}),
    ("FairletClustering", {"seed":42, "n_clusters": 2}),
    ("VariationalFairClustering", {"seed":42, "n_clusters": 2}),
])
def test_clustering_inprocessor(mitigator_name, mitigator_params, clustering_dataset):
    metrics1 = run_inprocessing_clustering(clustering_dataset, mitigator_name, mitigator_params)
    metrics2 = run_inprocessing_clustering_pipeline(clustering_dataset, mitigator_name, mitigator_params)
    check_results(metrics1, metrics2)



@pytest.mark.parametrize("mitigator_name, mitigator_params", [
    ("PopularityPropensityMF", {'K':40, 'beta': 0.02, 'steps':10}),
    ("BlindSpotAwareMF", {'K':40, 'beta': 0.02, 'steps':10}),
    ("FairRec", {'rec_size':10, 'MMS_fraction':0.5}),
    ("DebiasingLearningMF", {'K':40, 'normalization':'Vanilla', 'lamda':0.08, 'metric':'mse', 'bias_mode':'Regularized', 'seed':1}),
])
def test_recsys_inprocessor(mitigator_name, mitigator_params, recommender_dataset):
    metrics1 = run_inprocessing_recsys(recommender_dataset, recommender_bias_metrics, mitigator_name, mitigator_params)
    metrics2 = run_inprocessing_recsys_pipeline(recommender_dataset, recommender_bias_metrics, mitigator_name, mitigator_params)
    check_results(metrics1, metrics2)


def run_inprocessing_categorical(dataset, bias_metrics, estimator_class, mitigator_name, model_params, mitigator_params, is_multiclass=False):
    train = dataset['train']
    test = dataset['test']

    model = estimator_class(**model_params)
    
    inp = get_inprocessor(mitigator_name, mitigator_params)
    inp.transform_estimator(model)
    inp.fit(X=train['X'], y=train['y'], group_a=train['group_a'], group_b=train['group_b'])
    y_pred = inp.predict(X=test['X'], group_a=test['group_a'], group_b=test['group_b'])
    
    if is_multiclass:
        return bias_metrics(test['group_a'],  y_pred, test['y'])
    else:
        return bias_metrics(test['group_a'], test['group_b'], y_pred, test['y'])


def run_inprocessing_categorical_pipeline(dataset, bias_metrics, estimator_class, mitigator_name, model_params, mitigator_params, is_multiclass=False):
    train = dataset['train']
    test = dataset['test']

    estimator = estimator_class(**model_params)
    inp = get_inprocessor(mitigator_name, mitigator_params)
    inp.transform_estimator(estimator)

    pipeline = Pipeline(
        steps=[
            ("bm_inprocessing", inp),
        ]
    )

    pipeline.fit(X=train['X'], y=train['y'], bm__group_a=train['group_a'], bm__group_b=train['group_b'])
    y_pred = pipeline.predict(X=test['X'], bm__group_a=test['group_a'], bm__group_b=test['group_b'])
    
    if is_multiclass:
        return bias_metrics(test['group_a'],  y_pred, test['y'])
    else:
        return bias_metrics(test['group_a'], test['group_b'], y_pred, test['y'])
    

def run_inprocessing_clustering(dataset, mitigator_name, mitigator_params):
    train = dataset['train']

    inp = get_inprocessor(mitigator_name, mitigator_params)
    fit_params = {
        "X": train['X'],
        "group_a": train['group_a'], 
        "group_b": train['group_b']
        }
    inp.fit(**fit_params)
    
    if mitigator_name == "FairKMedianClustering":
        y_pred = inp.labels_
    else:
        y_pred = inp.predict(train['X'], **fit_params)
    
    if mitigator_name == "FairKCenterClustering":
        centroids = inp.all_centroids
    else:
        centroids = inp.cluster_centers_

    return clustering_bias_metrics(train['group_a'], train['group_b'], y_pred, train['X'], centroids)


def run_inprocessing_clustering_pipeline(dataset, mitigator_name, mitigator_params):
    train = dataset['train']

    inp = get_inprocessor(mitigator_name, mitigator_params)

    pipeline = Pipeline(
        steps=[
            ("bm_inprocessing", inp),
        ]
    )

    fit_params = {
        "bm__group_a": train['group_a'], 
        "bm__group_b": train['group_b']
        }

    pipeline.fit(train['X'], **fit_params)

    pred_params = {
        "bm__group_a": train['group_a'], 
        "bm__group_b": train['group_b']
        }
    if mitigator_name == "FairKMedianClustering":
        y_pred = pipeline['bm_inprocessing'].labels_
    else:
        y_pred = pipeline.predict(X=train['X'], **pred_params)

    if mitigator_name == "FairKCenterClustering":
        centroids = pipeline['bm_inprocessing'].all_centroids
    else:
        centroids = pipeline['bm_inprocessing'].cluster_centers_

    return clustering_bias_metrics(train['group_a'], train['group_b'], y_pred, train['X'], centroids)


def run_inprocessing_recsys(dataset, bias_metrics, mitigator_name, mitigator_params):
    train = dataset['train']
    inp = get_inprocessor(mitigator_name, mitigator_params)
    data_matrix = train['data_pivot'].fillna(0).to_numpy()
    inp.fit(data_matrix)
    rankings  = inp.predict(data_matrix, top_n=10)
    rankings = rankings.astype({'score':'float64'})
    mat = rankings.pivot(columns='Y',index='X',values='score').replace(np.nan,0).to_numpy()

    return bias_metrics(mat_pred=mat>0, metric_type='item_based')


def run_inprocessing_recsys_pipeline(dataset, bias_metrics, mitigator_name, mitigator_params):
    train = dataset['train']
    inp = get_inprocessor(mitigator_name, mitigator_params)
    data_matrix = train['data_pivot'].fillna(0).to_numpy()

    pipeline = Pipeline(
        steps=[
            ("bm_inprocessing", inp),
        ]
    )
    pipeline.fit(data_matrix)
    rankings  = pipeline.predict(data_matrix, top_n=10)
    rankings = rankings.astype({'score':'float64'})
    mat = rankings.pivot(columns='Y',index='X',values='score').replace(np.nan,0).to_numpy()

    return bias_metrics(mat_pred=mat>0, metric_type='item_based')


from collections import defaultdict
from itertools import product

from holisticai.bias.mitigation.commons.disparate_impact_remover._categorical_feature import CategoricalFeature
from holisticai.bias.mitigation.commons.disparate_impact_remover._utils import (
    assign_overflow,
    flow_on_group_features,
    gen_desired_count,
    gen_desired_dist,
    get_categories_count,
    get_categories_count_norm,
    get_group_data,
    get_median,
    get_median_per_category,
    get_mode,
)


class CategoricalRepairer:
    def __init__(self, feature_to_repair, repair_level, kdd=False, features_to_ignore=None):
        """
        Initializes a CategoricalRepairer object.

        Parameters
        ----------

        feature_to_repair: str
            The name of the feature to repair.
        repair_level: float
            The desired repair level, ranging from 0 to 1.
        kdd: (bool, optional)
            Whether to use the KDD repair method. Defaults to False.
        features_to_ignore: (list, optional)
            A list of features to ignore during repair. Defaults to None.
        """
        self.feature_to_repair = feature_to_repair
        self.repair_level = repair_level
        self.kdd = kdd
        self.features_to_ignore = [] if features_to_ignore is None else features_to_ignore
        self.preprocessor = DataPreprocessor(self.feature_to_repair, self.kdd, self.features_to_ignore)

    def repair(self, data_to_repair):
        """
        Repairs the given data by applying bias mitigation techniques.

        Parameters
        ----------
        data_to_repair: list
            The data to be repaired.

        Returns:
            list: The repaired data.
        """
        (
            col_ids,
            cols_to_repair,
            data_dict,
            repair_types,
            unique_col_vals,
            index_lookup,
            all_stratified_groups,
            stratified_group_indices,
            val_sets,
            stratified_group_data,
        ) = self.preprocessor.preprocessing(data_to_repair)

        mode_feature_to_repair = get_mode(data_dict[self.feature_to_repair])

        for col_id in cols_to_repair:
            col = data_dict[col_id]
            repair_type = repair_types[col_id]

            if repair_type in {int, float}:
                NumericRepairer(self.repair_level, self.kdd).repair(
                    col,
                    data_dict,
                    col_id,
                    unique_col_vals,
                    index_lookup,
                    all_stratified_groups,
                    stratified_group_data,
                    val_sets,
                )

            elif repair_type in {str}:
                StringRepairer(self.feature_to_repair, self.repair_level).repair(
                    col,
                    data_dict,
                    col_id,
                    all_stratified_groups,
                    stratified_group_indices,
                    stratified_group_data,
                    mode_feature_to_repair,
                )

        repaired_data = []
        for i, orig_row in enumerate(data_to_repair):
            new_row = [orig_row[j] if j not in cols_to_repair else data_dict[j][i] for j in col_ids]
            repaired_data.append(new_row)

        return repaired_data


class DataPreprocessor:
    def __init__(self, feature_to_repair, kdd, features_to_ignore):
        """
        Initializes a CategoricalRepairer object.

        Parameters
        ----------
        feature_to_repair: str
            The name of the feature to repair.
        kdd: float
            The KDD value used for repairing the feature.
        features_to_ignore: list
            A list of features to ignore during the repair process.
        """
        self.feature_to_repair = feature_to_repair
        self.kdd = kdd
        self.features_to_ignore = features_to_ignore

    def preprocessing(self, data_to_repair):
        """
        Preprocesses the data for repair.

        Parameters
        ----------

        data_to_repair: list
            The data to be repaired.
        """
        num_cols = len(data_to_repair[0])
        col_ids = list(range(num_cols))

        col_types = [
            "Y" if i not in self.features_to_ignore else "I" if i != self.feature_to_repair else "X" for i in col_ids
        ]
        col_type_dict = dict(zip(col_ids, col_types))
        not_I_col_ids = [x for x in col_ids if col_type_dict[x] != "I"]

        cols_to_repair = (
            [x for x in col_ids if col_type_dict[x] in "YX"]
            if not self.kdd
            else [x for x in col_ids if col_type_dict[x] == "Y"]
        )

        safe_stratify_cols = [self.feature_to_repair]

        data_dict = {col_id: [] for col_id in col_ids}
        for row in data_to_repair:
            for i in col_ids:
                data_dict[i].append(row[i])

        repair_types = {}
        for col_id, values in list(data_dict.items()):
            if all(isinstance(value, float) for value in values):
                repair_types[col_id] = float
            elif all(isinstance(value, int) for value in values):
                repair_types[col_id] = int
            else:
                repair_types[col_id] = str

        unique_col_vals = {}
        index_lookup = {}
        for col_id in not_I_col_ids:
            col_values = data_dict[col_id]
            col_values = sorted(set(col_values))
            unique_col_vals[col_id] = col_values
            col_values_len = len(col_values)
            index_lookup[col_id] = {col_values[i]: i for i in range(col_values_len)}

        unique_stratify_values = [unique_col_vals[i] for i in safe_stratify_cols]

        stratified_group_indices = defaultdict(list)
        val_sets = defaultdict(lambda: defaultdict(set))
        for i, row in enumerate(data_to_repair):
            group = tuple(row[col] for col in safe_stratify_cols)
            for col_id in cols_to_repair:
                val_sets[group][col_id].add(row[col_id])
            stratified_group_indices[group].append(i)

        all_stratified_groups = list(product(*unique_stratify_values))

        stratified_group_data = StratifiedGroupDataComputer().compute(
            data_dict, stratified_group_indices, all_stratified_groups
        )

        return (
            col_ids,
            cols_to_repair,
            data_dict,
            repair_types,
            unique_col_vals,
            index_lookup,
            all_stratified_groups,
            stratified_group_indices,
            val_sets,
            stratified_group_data,
        )


class StratifiedGroupDataComputer:
    def compute(self, data_dict, stratified_group_indices, all_stratified_groups):
        """
        Computes the stratified group data based on the given data dictionary, stratified group indices, and all stratified groups.

        Paramters
        ---------

        data_dict: dict
            A dictionary containing the data for each column.
        stratified_group_indices: dict
            A dictionary containing the indices of each stratified group.
        all_stratified_groups: list
            A list of all stratified groups.

        Returns:
            dict: A dictionary containing the stratified group data for each group and column.
        """
        stratified_group_data = defaultdict(dict)
        for group in all_stratified_groups:
            for col_id, col_dict in data_dict.items():
                indices = defaultdict(list)
                for i in stratified_group_indices[group]:
                    value = col_dict[i]
                    indices[value].append(i)
                stratified_col_values = sorted((occurs, val) for val, occurs in indices.items())
                stratified_col_values.sort(key=lambda tup: tup[1])
                stratified_group_data[group][col_id] = stratified_col_values
        return stratified_group_data


class StringRepairer:
    def __init__(self, feature_to_repair, repair_level):
        """
        Initialize the CategoricalRepairer object.

        Parameters
        ---------

        feature_to_repair: str
            The name of the feature to repair.
        repair_level: float
            The repair level to be applied to the feature.
        """
        self.feature_to_repair = feature_to_repair
        self.repair_level = repair_level

    def repair(
        self,
        col,
        data_dict,
        col_id,
        all_stratified_groups,
        stratified_group_indices,
        stratified_group_data,
        mode_feature_to_repair,
    ):
        """
        Repairs the categorical feature values in the given column based on the provided data.

        Parameters
        ----------

        col: str
            The name of the column to repair.
        data_dict: dict
            A dictionary containing the data.
        col_id: int
            The ID of the column to repair.
        all_stratified_groups: list
            A list of all stratified groups.
        stratified_group_indices: dict
            A dictionary mapping each stratified group to its corresponding indices.
        stratified_group_data: dict
            A dictionary containing the data for each stratified group.
        mode_feature_to_repair: str
            The mode feature to repair.
        """
        feature = CategoricalFeature(col)
        categories = list(feature.bin_index_dict.keys())

        group_features = get_group_data(all_stratified_groups, stratified_group_data, col_id)
        categories_count = get_categories_count(categories, all_stratified_groups, group_features)
        categories_count_norm = get_categories_count_norm(
            categories, all_stratified_groups, categories_count, group_features
        )
        median = get_median_per_category(categories, categories_count_norm)

        def dist_generator(group_index, category):
            return gen_desired_dist(
                group_index,
                category,
                col_id,
                median,
                self.repair_level,
                categories_count_norm,
                self.feature_to_repair,
                mode_feature_to_repair,
            )

        def count_generator(group_index, group, category):
            return gen_desired_count(
                group_index, group, category, median, group_features, self.repair_level, categories_count
            )

        group_features, overflow = flow_on_group_features(all_stratified_groups, group_features, count_generator)
        group_features, _, _ = assign_overflow(
            all_stratified_groups, categories, overflow, group_features, dist_generator
        )

        repaired_values = {group: group_features[group].data for group in all_stratified_groups}

        for group in all_stratified_groups:
            indices = stratified_group_indices[group]
            group_repaired_values = repaired_values[group]
            for i, index in enumerate(indices):
                data_dict[col_id][index] = group_repaired_values[i]


class NumericRepairer:
    def __init__(self, repair_level, kdd):
        """
        Initializes a new instance of the CategoricalRepairer class.

        Parameters
        ----------

        repair_level: float
            The repair level to be applied.
        kdd: bool
            A flag indicating whether to use KDD repair or not.
        """
        self.repair_level = repair_level
        self.kdd = kdd

    def repair(
        self,
        col,
        data_dict,
        col_id,
        unique_col_vals,
        index_lookup,
        all_stratified_groups,
        stratified_group_data,
        val_sets,
    ):
        """
        Repairs the categorical column values based on the provided parameters.

        Parameters
        ----------

        col: list
            The list of column values.
        data_dict: dict
            The dictionary containing the data.
        col_id: int
            The ID of the column to be repaired.
        unique_col_vals: dict
            The dictionary containing unique column values.
        index_lookup: dict
            The dictionary containing the index lookup.
        all_stratified_groups: list
            The list of all stratified groups.
        stratified_group_data: dict
            The dictionary containing stratified group data.
        val_sets: dict
            The dictionary containing validation sets.

        Returns:
            None
        """
        num_quantiles = min(len(val_sets[group][col_id]) for group in all_stratified_groups)
        quantile_unit = 1.0 / num_quantiles
        group_offsets = {group: 0 for group in all_stratified_groups}
        quantiles_data = []

        for _ in range(num_quantiles):
            median_at_quantiles = []
            indices_per_group = {}

            for group in all_stratified_groups:
                group_data_at_col = stratified_group_data[group][col_id]
                num_vals = len(group_data_at_col)
                group_offset = group_offsets[group] * num_vals
                offset = int(round(group_offset))
                number_to_get = int(round(group_offset + quantile_unit * num_vals)) - offset
                group_offsets[group] += quantile_unit

                if number_to_get > 0:
                    offset_data = group_data_at_col[offset : offset + number_to_get]
                    indices_per_group[group] = [i for val_indices, _ in offset_data for i in val_indices]
                    values = sorted(float(val) for _, val in offset_data)
                    median_at_quantiles.append(get_median(values, self.kdd))

            median = get_median(median_at_quantiles, self.kdd)
            quantiles_data.append((median, indices_per_group))

        median_val_pos = {q[0]: index_lookup[col_id][q[0]] for q in quantiles_data}

        for median, indices_per_group in quantiles_data:
            median_pos = median_val_pos[median]
            for group in all_stratified_groups:
                indices_to_repair = indices_per_group.get(group, [])
                if not indices_to_repair:
                    continue

                original_values = [col[index] for index in indices_to_repair]
                current_val_positions = [index_lookup[col_id][value] for value in original_values]
                distances_to_repair = [
                    int(round((median_pos - position) * self.repair_level)) for position in current_val_positions
                ]
                indices_of_repair_values = [
                    current_val_positions[i] + distances_to_repair[i] for i in range(len(indices_to_repair))
                ]
                repaired_values = [unique_col_vals[col_id][index] for index in indices_of_repair_values]

                for i in range(len(indices_to_repair)):
                    data_dict[col_id][indices_to_repair[i]] = repaired_values[i]



from __future__ import annotations

from typing import Optional

import numpy as np
import pandas as pd


class GridGenerator:
    """
    Create a grid using a constraint
    """

    def __init__(
        self,
        grid_size: Optional[int] = 5,
        grid_limit: Optional[float] = 2.0,
        neg_allowed: Optional[np.ndarray] = None,
        force_L1_norm: Optional[bool] = True,
    ):
        """
        Initialize Grid Generator

        Parameters
        ----------
        grid_size: int
            Number of columns to be generated in the grid.

        grid_limit: float
            Range of the values in the grid generated.

        neg_allowed: np.ndarray
            Ensures if we want to include negative values in the grid or not.
            If True, the range is doubled.

        force_L1_norm: bool
            Ensures L1 normalization if set to True.
        """
        self.grid_size = grid_size
        self.grid_limit = grid_limit
        self.neg_allowed = neg_allowed if neg_allowed is not None else np.zeros((self.grid_size,), dtype=bool)
        self.force_L1_norm = force_L1_norm

    def generate_grid(self, constraint):
        """
        Generate a grid of lambda vectors based on the provided constraint.

        Parameters
        ----------
        constraint: Constraint
            Constraint object containing the basis for grid generation.

        Returns
        -------
        grid: pd.DataFrame
            Generated grid of lambda vectors.
        """
        self.dim = len(constraint.basis["+"].columns)
        coefs = self._generate_coefs()
        grid = constraint.basis["+"].dot(coefs["+"]) + constraint.basis["-"].dot(coefs["-"])
        return grid

    def _generate_coefs(self):
        np_grid_values = self._build_grid()
        grid_values = pd.DataFrame(np_grid_values[: self.grid_size]).T
        pos_grid_values = grid_values.clip(lower=0)
        neg_grid_values = -grid_values.clip(upper=0)
        lambda_vector = {"+": pos_grid_values, "-": neg_grid_values}
        return lambda_vector

    def _build_grid(self):
        max_value = self._get_true_dim()
        self.accumulator = []

        while len(self.accumulator) < self.grid_size:
            entry = np.zeros(self.dim)
            self._accumulate_integer_grid(0, max_value, entry)
            max_value += 1

        xs = np.array(self.accumulator) * self.grid_limit / max_value
        return xs

    def _accumulate_integer_grid(self, index, max_val, entry):
        if index == self.dim:
            self.accumulator.append(entry.copy())
            return

        if (index == self.dim - 1) and self.force_L1_norm:
            values = [-max_val, max_val] if self.neg_allowed[index] and max_val > 0 else [max_val]
        else:
            min_val = -max_val if self.neg_allowed[index] else 0
            values = range(min_val, max_val + 1)

        for current_value in values:
            entry[index] = current_value
            self._accumulate_integer_grid(index + 1, max_val - abs(current_value), entry)

    def _get_true_dim(self):
        true_dim = self.dim - 1 if self.force_L1_norm else self.dim
        n_units = (self.grid_size / (2 ** self.neg_allowed.sum())) ** (1 / true_dim) - 1
        return max(0, int(np.floor(n_units)))



"""
The :mod:`holisticai.utils` module includes utils helper tools
"""

# formatting
from holisticai.utils._definitions import (
    BinaryClassificationProxy,
    ConditionalImportances,
    Importances,
    LocalConditionalImportances,
    LocalImportances,
    ModelProxy,
    MultiClassificationProxy,
    PartialDependence,
    RegressionProxy,
    create_proxy,
)
from holisticai.utils._formatting import (
    extract_columns,
    extract_group_vectors,
    mat_to_binary,
    normalize_tensor,
    recommender_formatter,
)

# plotting
from holisticai.utils._plotting import get_colors
from holisticai.utils.inspection._partial_dependence import compute_partial_dependence

__all__ = [
    "extract_columns",
    "mat_to_binary",
    "normalize_tensor",
    "get_colors",
    "recommender_formatter",
    "extract_group_vectors",
    "BinaryClassificationProxy",
    "MultiClassificationProxy",
    "RegressionProxy",
    "compute_partial_dependence",
    "create_proxy",
    "Importances",
    "LocalImportances",
    "LocalConditionalImportances",
    "PartialDependence",
    "ConditionalImportances",
    "ModelProxy",
]



from __future__ import annotations

from typing import Union

import numpy as np
import pandas as pd
from numpy.random import RandomState
from sklearn.metrics import accuracy_score, mean_squared_error

from holisticai.utils import ConditionalImportances, Importances, ModelProxy
from holisticai.utils.feature_importances import group_samples_by_learning_task

metric_scores = {
    "binary_classification": accuracy_score,
    "regression": mean_squared_error,
    "multi_classification": accuracy_score,
}


def compute_surrogate_feature_importance(
    proxy: ModelProxy,
    X: pd.DataFrame,
    y: Union[pd.Series, None] = None,
    random_state: Union[RandomState, int, None] = None,
    conditional: bool = False,
) -> Union[Importances, ConditionalImportances]:
    if conditional and y is None:
        raise ValueError("y must be provided when conditional=True")

    pfi = SurrogateFeatureImportanceCalculator(random_state=random_state)
    if conditional:
        sample_groups = group_samples_by_learning_task(y, proxy.learning_task)
        values = {
            group_name: pfi.compute_importances(X=X.loc[indexes], proxy=proxy)
            for group_name, indexes in sample_groups.items()
        }
        return ConditionalImportances(values=values)
    return pfi.compute_importances(X=X, proxy=proxy)


class SurrogateFeatureImportanceCalculator:
    def __init__(self, random_state: Union[RandomState, int, None] = None, importance_type: str = "global"):
        if random_state is None:
            random_state = RandomState(42)
        self.random_state = random_state
        self.importance_type = importance_type

    def create_surrogate_model(self, X: pd.DataFrame, y: pd.Series, learning_task: str):
        if learning_task in ["binary_classification", "multi_classification"]:
            from sklearn.tree import DecisionTreeClassifier

            dt = DecisionTreeClassifier(max_depth=3, random_state=self.random_state)
            return dt.fit(X, y)

        if learning_task == "regression":
            from sklearn.tree import DecisionTreeRegressor

            dt = DecisionTreeRegressor(max_depth=3, random_state=self.random_state)
            return dt.fit(X, y)

        raise ValueError("model_type must be either 'binary_classification', 'multi_classification' or 'regression'")

    def compute_importances(self, X: pd.DataFrame, proxy: ModelProxy) -> Importances:
        """
        Compute surrogate feature importance for a given model type, model and input features.

        Parameters
        ----------
            x (pandas.DataFrame): The input features.
            proxy ModelProxy: The model proxy.

        Returns:
            holisticai.utils.feature_importance.SurrogateFeatureImportance: The surrogate feature importance.
        """
        y_pred = proxy.predict(X)
        surrogate = self.create_surrogate_model(X, y_pred, proxy.learning_task)
        feature_names = X.columns
        importances = surrogate.feature_importances_
        feature_importances = pd.DataFrame.from_dict(
            {"Variable": feature_names, "Importance": importances}
        ).sort_values("Importance", ascending=False)
        feature_importances["Importance"] = feature_importances["Importance"] / feature_importances["Importance"].sum()

        feature_names = list(feature_importances["Variable"].values)
        importances = np.array(feature_importances["Importance"].values)
        return Importances(values=importances, feature_names=feature_names, extra_attrs={"surrogate": surrogate})



"""
The :mod:`holisticai.datasets.plots` module includes bias plotters.
"""

from holisticai.datasets.plots._exploratory import (
    correlation_matrix_plot,
)

# All bias plotters
__all__ = [
    "correlation_matrix_plot",
]



import os
import sys

sys.path.append(os.getcwd())
import numpy as np
import pytest
from utils import (
    numeric_config,
    numeric_nan_config,
    str_config,
    unconventional_numeric_config,
)

from holisticai.bias.metrics import confusion_matrix

##################################################################################################################
# 2 inputs
##################################################################################################################
# String and unconventional y values

metrics_functions = [confusion_matrix]


@pytest.mark.parametrize("metric_function", metrics_functions)
def test_string_y_values(metric_function):
    """metric must support string labels"""
    _, y_pred, y_true = str_config()
    metric_function(y_pred, y_true)


@pytest.mark.parametrize("metric_function", metrics_functions)
def test_unconventional_y_values(metric_function):
    """metric support unconventional numeric labels"""
    _, y_pred, y_true = unconventional_numeric_config()
    metric_function(y_pred, y_true)


@pytest.mark.xfail()
@pytest.mark.parametrize("metric_function", metrics_functions)
def test_raise_invalid_y_values(metric_function):
    """metric must validate y values in classes"""
    _, y_pred, y_true = numeric_config()
    classes = [2, 5]
    metric_function(y_pred, y_true, classes=classes)


@pytest.mark.xfail(raises=ValueError)
@pytest.mark.parametrize("metric_function", metrics_functions)
def test_y_pred_nan(metric_function):
    """metric must validate non nan values in y_pred"""
    _, y_pred, y_true = numeric_nan_config(y_pred_nan=True)
    metric_function(y_pred, y_true)


@pytest.mark.xfail(raises=ValueError)
@pytest.mark.parametrize("metric_function", metrics_functions)
def test_y_true_nan(metric_function):
    """metric must validate non nan values in y_true"""
    _, y_pred, y_true = numeric_nan_config(y_pred_nan=False)
    metric_function(y_pred, y_true)



from holisticai.datasets import load_dataset
from holisticai.security.commons import DataMinimizer
from sklearn.compose import ColumnTransformer
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import OneHotEncoder, StandardScaler
from sklearn.tree import DecisionTreeClassifier
from holisticai.utils import BinaryClassificationProxy

dataset = load_dataset('adult')

train_test = dataset.train_test_split(0.2, random_state=42)
train = train_test['train']
test = train_test['test']

# Identify categorical and numerical features
categorical_features = train['X'].select_dtypes(include=['category']).columns
numerical_fatures = train['X'].select_dtypes(exclude=['category']).columns

# Create transformers for numerical and categorical features
numeric_transformer = Pipeline(steps=[('scaler', StandardScaler())])
categorical_transformer = Pipeline(steps=[('onehot', OneHotEncoder(handle_unknown='ignore'))])

# Combine transformers into a preprocessor using ColumnTransformer
preprocessor = ColumnTransformer(
    transformers=[
        ('num', numeric_transformer, numerical_fatures),
        ('cat', categorical_transformer, categorical_features)
])


model = Pipeline(steps=[('preprocessor', preprocessor),
                        ('classifier', DecisionTreeClassifier())])

model.fit(train['X'], train['y'])

y_pred = model.predict(test['X'])
y_proba = model.predict_proba(test['X'])

print('Base model accuracy: ', model.score(test['X'], test['y']))

proxy = BinaryClassificationProxy(predict=model.predict,
                                    predict_proba=model.predict_proba,
                                    classes=[0, 1])

entry = DataMinimizer(proxy=proxy)

entry.fit(train['X'], train['y'])


import numpy as np
from holisticai.bias.mitigation.postprocessing.plugin_estimator_and_recalibration.algorithm_utils import f_lambda
from holisticai.utils.transformers.bias import SensitiveGroups


class PluginEstimationAndCalibrationAlgorithm:
    """Plugin Estimation and Calibration Algorithm

    This class implements the Plugin Estimation and Calibration Algorithm (PECA) for bias mitigation.
    This algorithm was proposed by Hardt et al. in their paper titled "Fair Regression via Plug-In Estimator and Recalibration".
    The paper can be found at: https://proceedings.neurips.cc/paper_files/paper/2020/file/ddd808772c035aed516d42ad3559be5f-Paper.pdf

    References
    ----------
    .. [1] Chzhen, E., Denis, C., Hebiri, M., Oneto, L., & Pontil, M. (2020).
        Fair regression via plug-in estimator and recalibration with statistical guarantees.
        Advances in Neural Information Processing Systems, 33, 19137-19148.
    """

    def __init__(self, L=25, beta=0.1):
        """Parameters
        ----------
        length : int, optional (default=25)
            The length of the range of values to consider for the output predictions.
            The range of values is [-length, length].
        beta : float, optional (default=0.1)
            The value of the beta parameter used in the calculation of the lambda values.

        Attributes
        ----------
        multiplier : int
            The multiplier used to convert the output predictions to the range [0, 1].
        length : int
            The length of the range of values to consider for the output predictions.
            The range of values is [-length, length].
        beta : float
            The value of the beta parameter used in the calculation of the lambda values.
        sensitive_groups : SensitiveGroups
            The SensitiveGroups object used to transform the sensitive features.
        epsilon : float
            The value of the epsilon parameter used to avoid division by zero.
        probabilities : list
            The probabilities for each sensitive group.
        lambda_values : np.ndarray
            The lambda values used in the calculation of the output predictions.
            self.multiplier = 1
            self.length = np.floor(L / 2).astype(np.int32)
            self.beta = beta
            self.sensitive_groups = SensitiveGroups()
            self.epsilon = np.finfo(float).eps
        """
        self.multiplier = 1
        self.length = np.floor(L / 2).astype(np.int32)
        self.beta = beta
        self.sensitive_groups = SensitiveGroups()
        self.epsilon = np.finfo(float).eps

    def fit(self, y_pred: np.ndarray, sensitive_features: np.ndarray):
        # Fit and transform the sensitive features
        transformed_sensitive_features = self.sensitive_groups.fit_transform(sensitive_features, convert_numeric=True)

        # Calculate the adjusted predictions
        adjusted_predictions = y_pred * 2 - 1

        # Calculate the probabilities for each sensitive group
        self.probabilities = np.array(
            [
                np.mean(transformed_sensitive_features == 0),
                np.mean(transformed_sensitive_features == 1),
            ]
        )

        # Compute the lambda values
        self.lambda_values = f_lambda(
            adjusted_predictions,
            transformed_sensitive_features,
            self.multiplier,
            self.length,
            self.beta,
        )

    def transform(self, y_pred: np.ndarray, sensitive_features: np.ndarray):
        transformed_sensitive_features = self.sensitive_groups.transform(sensitive_features, convert_numeric=True)
        adjusted_predictions = y_pred * 2 - 1
        index_range = np.arange(-self.length, self.length + 1, 1)

        transformed_sensitive_features = transformed_sensitive_features.to_list()
        minimizing_values = (
            np.expand_dims(self.probabilities[transformed_sensitive_features], axis=1)
            * np.square(np.expand_dims(adjusted_predictions, axis=1) - index_range * self.multiplier / self.length)
            + (1 - 2 * np.expand_dims(transformed_sensitive_features, axis=1)) * self.lambda_values
        )
        min_indices = np.argmin(minimizing_values, axis=1)
        output_predictions = index_range[min_indices] * self.multiplier / self.length
        output_predictions = (output_predictions + 1) / 2
        return output_predictions



import numpy as np
from holisticai.utils import Importances

def test_compute_ranked_feature_importance():
    # Create a sample feature importance dataframe
    feature_importance = Importances(feature_names=['feature1', 'feature2', 'feature3', 'feature4'], values=np.array([0.6, 0.15, 0.15, 0.1]))

    # Test filtering with default threshold (alpha=None)
    ranked_importances = feature_importance.top_alpha()
    assert np.isclose(ranked_importances.values , np.array([0.6, 0.15])).all()
    assert ranked_importances.feature_names == ['feature1', 'feature2']

    # Test filtering with custom threshold (alpha=0.8)
    ranked_importances = feature_importance.top_alpha(alpha=0.8)
    assert np.isclose(ranked_importances.values, np.array([0.6, 0.15])).all()
    assert ranked_importances.feature_names == ['feature1', 'feature2']

    # Test filtering with custom threshold (alpha=0.5)
    ranked_importances = feature_importance.top_alpha(alpha=0.5)
    assert np.isclose(ranked_importances.values , np.array([0.6])).all()
    assert ranked_importances.feature_names == ['feature1']

    # Test filtering with custom threshold (alpha=0.2)
    ranked_importances = feature_importance.top_alpha(alpha=0.2)
    assert np.isclose(ranked_importances.values , np.array([0.6])).all()
    assert ranked_importances.feature_names == ['feature1']

def test_importances_object():
    values = np.array([0.50, 0.30, 0.20])
    feature_names = ["feature_1", "feature_2", "feature_3"]
    feature_importance = Importances(values=values, feature_names=feature_names)
    ranked_feature_importance = feature_importance.top_alpha(alpha=0.8)
    print(ranked_feature_importance)


from __future__ import annotations

from typing import Literal, Optional

import numpy as np
from holisticai.bias.mitigation.inprocessing.commons.classification import _constraints as cc
from holisticai.bias.mitigation.inprocessing.commons.regression import _constraints as rc
from holisticai.bias.mitigation.inprocessing.commons.regression import _losses as rl
from holisticai.bias.mitigation.inprocessing.exponentiated_gradient.algorithm import ExponentiatedGradientAlgorithm
from holisticai.utils.transformers.bias import BMInprocessing as BMImp
from sklearn.base import BaseEstimator, ClassifierMixin, clone


class ExponentiatedGradientReduction(BaseEstimator, ClassifierMixin, BMImp):
    """Exponentiated gradient reduction

    Exponentiated gradient reduction [1]_ is an in-processing technique that reduces\
    fair classification to a sequence of cost-sensitive classification problems,\
    returning a randomized classifier with the lowest empirical error subject to\
    fair classification constraints.

    Parameters
    ----------

        estimator : sklearn-like
            The model you want to mitigate bias for.

        constraints: str
            The disparity constraints:
                - "DemographicParity"
                - "EqualizedOdds"
                - "TruePositiveRateParity"
                - "FalsePositiveRateParity"
                - "ErrorRateParity"

        eps: float
            Allowed fairness constraint violation; the solution is\
            guaranteed to have the error within ``2*best_gap`` of the best\
            error under constraint eps; the constraint violation is at most\
            ``2*(eps+best_gap)``.

        num_iter: int
            Maximum number of iterations.

        nu: float
            Convergence threshold for the duality gap, corresponding to a\
            conservative automatic setting based on the statistical\
            uncertainty in measuring classification error.

        eta_mul: float
            Initial setting of the learning rate.

        drop_prot_attr: bool
            Boolean flag indicating whether to drop protected\
            attributes from training data.

        loss : str
            String identifying loss function for constraints. Options include "ZeroOne", "Square", and "Absolute."

        min_val : float
            Loss function parameter for "Square" and "Absolute," typically the minimum of the range of y values.

        max_val: float
            Loss function parameter for "Square" and "Absolute," typically the maximum of the range of y values.

        verbose : int
            If >0, will show progress percentage.

        seed: int
            seed for random initialization

    References
    ---------
    .. [1] Agarwal, Alekh, et al. "A reductions approach to fair classification."
        International Conference on Machine Learning. PMLR, 2018.
    """

    CONSTRAINTS = Literal[
        "DemographicParity",
        "EqualizedOdds",
        "TruePositiveRateParity",
        "FalsePositiveRateParity",
        "ErrorRateParity",
    ]

    def __init__(
        self,
        constraints: str = "EqualizedOdds",
        eps: Optional[float] = 0.01,
        max_iter: Optional[int] = 50,
        nu: Optional[float] = None,
        eta0: Optional[float] = 2.0,
        loss: str = "ZeroOne",
        min_val: Optional[float] = None,
        max_val: Optional[float] = None,
        upper_bound: float = 0.01,
        verbose: Optional[int] = 0,
        estimator=None,
        seed: int = 0,
    ):
        self.constraints = constraints
        self.eps = eps
        self.max_iter = max_iter
        self.nu = nu
        self.eta0 = eta0
        self.loss = loss
        self.min_val = min_val
        self.max_val = max_val
        self.upper_bound = upper_bound
        self.verbose = verbose
        self.estimator = estimator
        self.seed = seed

    def transform_estimator(self, estimator):
        """
        This method is deprecated but retained for backwards-compatibility. You should pass the estimator object directly in the constructor.
        """

        self.estimator = estimator
        return self

    def fit(
        self,
        X: np.ndarray,
        y: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ):
        """
        Fit Exponentiated Gradient Reduction

        Parameters
        ----------
        X : matrix-like
            Input matrix
        y : array-like
            Target vector
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Returns
        -------
        Self

        """
        params = self._load_data(X=X, y=y, group_a=group_a, group_b=group_b)
        group_a = params["group_a"]
        group_b = params["group_b"]
        X = params["X"]
        y = params["y"]

        sensitive_features = np.stack([group_a, group_b], axis=1)

        self.estimator_ = clone(self.estimator)

        constraints_catalog = {
            "DemographicParity": cc.DemographicParity,
            "EqualizedOdds": cc.EqualizedOdds,
            "TruePositiveRateParity": cc.TruePositiveRateParity,
            "FalsePositiveRateParity": cc.FalsePositiveRateParity,
            "ErrorRateParity": cc.ErrorRateParity,
            "BoundedGroupLoss": rc.BoundedGroupLoss,
        }

        constraint_kargs = self._constraint_parameters()
        self.constraint_ = constraints_catalog[self.constraints](**constraint_kargs)

        self.model_ = ExponentiatedGradientAlgorithm(
            self.estimator_,
            constraints=self.constraint_,
            eps=self.eps,
            max_iter=self.max_iter,
            nu=self.nu,
            eta0=self.eta0,
            verbose=self.verbose,
            seed=self.seed,
        )

        self.model_.fit(X, y, sensitive_features=sensitive_features)

        return self

    def _constraint_parameters(self):
        kargs = {}
        if self.constraints == "BoundedGroupLoss":
            losses = {
                "ZeroOne": rl.ZeroOneLoss,
                "Square": rl.SquareLoss,
                "Absolute": rl.AbsoluteLoss,
            }
            if self.loss == "ZeroOne":
                self.loss_ = losses[self.loss]()
            else:
                self.loss_ = losses[self.loss](self.min_val, self.max_val)
            kargs.update({"loss": self.loss_, "upper_bound": self.upper_bound})
        return kargs

    def predict(self, X):
        """
        Prediction

        Description
        ----------
        Predict output for the given samples.

        Parameters
        ----------
        X : matrix-like
            Input matrix.

        Returns
        -------
        numpy.ndarray: Predicted output
        """
        return self.model_.predict(X).ravel()

    def predict_proba(self, X):
        """
        Predict Probabilities

        Description
        ----------
        Probability estimate for the given samples.

        Parameters
        ----------
        X : matrix-like
            Input matrix.

        Returns
        -------
        numpy.ndarray: probability output
        """
        return self.model_.predict_proba(X)



from __future__ import annotations

from holisticai.utils import Importances


def calculate_alpha_importance(feature_importance: Importances, alpha=0.8):
    """
    Calculates the alpha importance of a feature based on its feature importance.

    Parameters
    ----------
    feature_importance: list
        A list of feature importance values.
    alpha: (float, optional)
        The threshold value for filtering feature importance. Defaults to 0.8.

    Returns
    -------
        float: The alpha importance of the feature, calculated as the ratio of the number of features with importance
               greater than or equal to alpha to the total number of features.

    """
    alpha_feat_imp = feature_importance.top_alpha(alpha)
    len_alpha = len(alpha_feat_imp)
    len_100 = len(feature_importance)
    return len_alpha / len_100


class AlphaScore:
    """
    Represents the Fourth Fifths metric.
    """

    reference: int = 0
    name: str = "Alpha Importance Score"

    def __init__(self, alpha: float = 0.8):
        self.alpha = alpha

    def __call__(self, feature_importances: Importances):
        """
        Calculates the alpha importance of feature importance values.

        Parameters
        ----------
        feature_importance: Importances
            The feature importance values.

        Returns:
            float: The alpha importance value.
        """
        ranked_feature_importances = feature_importances.top_alpha(alpha=self.alpha)
        len_alpha = len(ranked_feature_importances)
        len_100 = len(feature_importances)
        return len_alpha / len_100


def alpha_score(feature_importance: Importances, alpha: float = 0.8):
    """
    Alpha Score calculates the proportion of features that account for the alpha percentage of the overall feature importance.

    Parameters
    ----------
    feature_importance: Importances
        The feature importance values.

    alpha: float, optional
        The alpha value represents the percentage of importance that will be considered in the calculation.
        For example, if alpha=0.8, the top 80% of the most important features will be considered.

    Returns
    -------
    float
        The alpha importance score

    Examples
    --------
    >>> from holisticai.explainability.commons import Importance
    >>> from holisticai.explainability.metrics import alpha_score
    >>> values = np.array([0.50, 0.30, 0.20])
    >>> feature_names = ["feature_1", "feature_2", "feature_3"]
    >>> feature_importance = Importance(values=values, feature_names=feature_names)
    >>> alpha_score(feature_importance)
    0.6666666666666666
    """

    metric = AlphaScore(alpha=alpha)
    return metric(feature_importance)



from __future__ import annotations

from typing import Literal

import numpy as np
from holisticai.utils.transformers.bias import BMPostprocessing as BMPost
from scipy.optimize import linprog
from sklearn.metrics import confusion_matrix


class EqualizedOdds(BMPost):
    """
    Equalized odds postprocessing use linear programming to find the probability with which change\
    favorable labels (y=1) to unfavorable labels (y=0) in the output estimator to optimize equalized odds.

    Parameters
    ----------
    solver : str
        Algorithm name used to solve the standard form problem. Solver supported must depend of your scipy poackage version.
        for scipy 1.9.0 the solvers available are:
        ["highs", "highs-ds", "highs-ipm", "interior-point", "revised simplex", "simplex"]

    seed : int
        Random seed for repeatability.

    References
    ----------
        .. [1] Hardt, Moritz, Eric Price, and Nati Srebro. "Equality of opportunity in supervised learning."\
        Advances in neural information processing systems 29 (2016).

        .. [2] Pleiss, Geoff, et al. "On fairness and calibration."\
        Advances in neural information processing systems 30 (2017).
    """

    SOLVERS = Literal[
        "highs",
        "highs-ds",
        "highs-ipm",
        "interior-point",
        "revised simplex",
        "simplex",
    ]

    def __init__(self, solver: str | None = "highs", seed: int | None = 42):
        self.solver = solver
        self.seed = seed

    def _objective_function_parameters_by_group(self, labels, predictions, group):
        # Compute Equalized Odds algorithm parameters for specific group

        labels = labels[group]
        predictions = predictions[group]

        prev = predictions
        flip_prev = 1 - predictions

        flip_predictions = 1 - predictions
        flip_labels = 1 - labels
        base_rate = np.mean(labels)

        tnr, fpr, fnr, tpr = confusion_matrix(labels, predictions, normalize="true").ravel()

        m_tn = np.logical_and(flip_predictions, flip_labels)
        m_fn = np.logical_and(flip_predictions, labels)
        m_fp = np.logical_and(predictions, flip_labels)
        m_tp = np.logical_and(predictions, labels)

        c = np.array([fpr - tpr, tnr - fnr])

        true_positive = (np.mean(prev * m_tp) - np.mean(flip_prev * m_tp)) / base_rate
        false_negative = (np.mean(flip_prev * m_fn) - np.mean(prev * m_fn)) / base_rate
        false_positive = (np.mean(prev * m_fp) - np.mean(flip_prev * m_fp)) / (1 - base_rate)
        true_negative = (np.mean(flip_prev * m_tn) - np.mean(prev * m_tn)) / (1 - base_rate)

        A_eq = np.array([[true_positive, false_negative], [false_positive, true_negative]])

        b_tp_fn = (np.mean(flip_prev * m_tp) + np.mean(prev * m_fn)) / base_rate
        b_fp_tn = (np.mean(flip_prev * m_fp) + np.mean(prev * m_tn)) / (1 - base_rate)

        b_eq = np.array([b_tp_fn, b_fp_tn])

        return {"c": c, "A_eq": A_eq, "b_eq": b_eq}

    def _build_objective_function(self, params_group_a, params_group_b):
        """
        Concat parameters from group a and b to create the objective function

        Parameters
        ----------
        params_group_a : dict
            dictionary of parameters with A_eq, b_eq, C keys.
        params_group_b : dict
            dictionary of parameters with A_eq, b_eq, C keys.

        Returns
        -------
        Tuple of objetive function parameters (A_eq, b_eq, C)
        """
        A_eq = np.concatenate([params_group_a["A_eq"], -params_group_b["A_eq"]], axis=1)
        b_eq = params_group_b["b_eq"] - params_group_a["b_eq"]
        C = np.concatenate([params_group_a["c"], params_group_b["c"]], axis=0)
        return A_eq, b_eq, C

    def _adjust_fairness(self, predictions, likelihoods, n2p, p2p):
        """
        Invert predictions and likelihoods with some probabilities

        Description
        ----------
        Use n2p and (1 - p2p) to invert output labels

        Parameters
        ----------
        predictions : numpy array
            binary predicted vector
        likelihoods : numpy array
            predicted score vector
        n2p : numpy array
            probability vector
        p2p : numpy array
            probability vector

        Returns
        -------
        numpy array
            new prediction vector with some inverted values

        numpy array
            new score vector with some inverted values
        """

        fair_predictions = predictions.copy()
        pos_indices = np.where(predictions == 1)[0]
        neg_indices = np.where(predictions == 0)[0]

        np.random.shuffle(pos_indices)
        np.random.shuffle(neg_indices)

        n2p_indices = neg_indices[: int(len(neg_indices) * n2p)]
        fair_predictions[n2p_indices] = 1

        p2n_indices = pos_indices[: int(len(pos_indices) * (1 - p2p))]
        fair_predictions[p2n_indices] = 0

        fair_likelihoods = likelihoods.copy()
        fair_likelihoods[n2p_indices] = 1 - fair_likelihoods[n2p_indices]
        fair_likelihoods[p2n_indices] = 1 - fair_likelihoods[p2n_indices]

        return fair_predictions, fair_likelihoods

    def fit(
        self,
        y: np.ndarray,
        y_pred: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ) -> EqualizedOdds:
        """
        Compute parameters for equalizing odds.

        Description
        ----------
        Build parameters for the objetive function and call the solver to find the algorithm parameters.

        Parameters
        ----------
        y : array-like
            Target vector (nb_examlpes,)
        y_pred : array-like
            Predicted vector (nb_examlpes,)
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Returns
        -------
            Self
        """
        if self.seed is not None:
            np.random.seed(self.seed)

        params = self._load_data(y=y, y_pred=y_pred, group_a=group_a, group_b=group_b)

        group_a = params["group_a"] == 1
        group_b = params["group_b"] == 1
        y = params["y"]
        y_pred = params["y_pred"]

        parameters_group_a = self._objective_function_parameters_by_group(y, y_pred, group_a)
        parameters_group_b = self._objective_function_parameters_by_group(y, y_pred, group_b)
        A_eq, b_eq, C = self._build_objective_function(parameters_group_a, parameters_group_b)

        A_ub = np.array(
            [
                [1, 0, 0, 0],
                [-1, 0, 0, 0],
                [0, 1, 0, 0],
                [0, -1, 0, 0],
                [0, 0, 1, 0],
                [0, 0, -1, 0],
                [0, 0, 0, 1],
                [0, 0, 0, -1],
            ],
            dtype=np.float64,
        )

        b_ub = np.array([1, 0, 1, 0, 1, 0, 1, 0], dtype=np.float64)

        # Linear program
        self.model_params = linprog(C, A_ub=A_ub, b_ub=b_ub, A_eq=A_eq, b_eq=b_eq, method=self.solver)

        return self

    def transform(self, y_pred: np.ndarray, group_a: np.ndarray, group_b: np.ndarray):
        """
        Apply transform function to predictions and likelihoods

        Description
        ----------
        Use the fitted probability to change the output label and invert the likelihood

        Parameters
        ----------
        y_pred : array-like
            Predicted vector (nb_examlpes,)
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Returns
        -------
        dict
            A dictionary with two keys, y_pred and y_score, which refers to the predicted labels and their probabilities, respectively.
        """
        params = self._load_data(y_pred=y_pred, group_a=group_a, group_b=group_b)

        group_a = params["group_a"] == 1
        group_b = params["group_b"] == 1
        y_pred = params["y_pred"]
        likelihoods = y_pred.copy().astype(np.float64)

        sp2p, sn2p, op2p, on2p = self.model_params.x

        predictions_a = y_pred[group_a]
        predictions_b = y_pred[group_b]
        likelihoods_a = likelihoods[group_a]
        likelihoods_b = likelihoods[group_b]

        # Randomly flip labels according to the probabilities in model_params
        fair_predictions_a, fair_likelihoods_a = self._adjust_fairness(predictions_a, likelihoods_a, sn2p, sp2p)
        fair_predictions_b, fair_likelihoods_b = self._adjust_fairness(predictions_b, likelihoods_b, on2p, op2p)

        # Mutated, fairer dataset with new labels4
        new_y_pred = y_pred.copy()
        new_y_pred[group_a] = fair_predictions_a
        new_y_pred[group_b] = fair_predictions_b

        new_y_score = likelihoods.copy()
        new_y_score[group_a] = fair_likelihoods_a
        new_y_score[group_b] = fair_likelihoods_b

        return {"y_pred": new_y_pred, "y_score": new_y_score}



from holisticai.security.commons._black_box_attack import BlackBoxAttack, classification_security_features
from holisticai.security.commons.data_minimization._core import DataMinimizer

__all__ = ["classification_security_features", "BlackBoxAttack", "DataMinimizer"]



import numpy as np
import pandas as pd
from sklearn.metrics import mean_squared_error


def rmse_score(y_true, y_pred, **kargs):
    return np.sqrt(mean_squared_error(y_true, y_pred, **kargs))


def smape(y_true, y_pred):
    return 1.0 / len(y_true) * np.sum(np.abs(y_pred - y_true) / (np.abs(y_true) + np.abs(y_pred)))


def regression_efficacy_metrics(y_pred, y_true):
    """
    Regression efficacy metrics batch computation.

    Description
    -----------
    This function computes all the relevant regression efficacy metrics,
    and displays them as a pandas dataframe. It also includes a reference value for comparison.

    Parameters
    ----------
    y_pred : array-like
        Predictions vector
    y_true (optional) : array-like
        Target vector

    Returns
    -------
    pandas DataFrame
        Metrics | Values | Reference
    """
    from sklearn import metrics

    perform = {
        # "Pseudo-R2": pr2_score,
        "RMSE": rmse_score,
        "MAE": metrics.mean_absolute_error,
        "MAPE": metrics.mean_absolute_percentage_error,
        "Max Error": metrics.max_error,
        "SMAPE": smape,
    }

    hyper = {
        # "Pseudo-R2": {},
        "RMSE": {"sample_weight": None, "multioutput": "uniform_average"},
        "MAE": {"sample_weight": None, "multioutput": "uniform_average"},
        "MAPE": {"sample_weight": None, "multioutput": "uniform_average"},
        "Max Error": {},
        "SMAPE": {},
    }

    ref_vals = {
        "RMSE": 0,
        "MAE": 0,
        "MAPE": 0,
        "Max Error": 0,
        "SMAPE": 0,
    }

    metrics = [[pf, fn(y_pred, y_true, **hyper[pf]), ref_vals[pf]] for pf, fn in perform.items()]

    return pd.DataFrame(metrics, columns=["Metric", "Value", "Reference"]).set_index("Metric")



import ipywidgets as widgets  # type: ignore
import matplotlib.pyplot as plt
from holisticai.bias.plots._bias_exploratory_plots import group_pie_plot, histogram_plot
from holisticai.bias.plots._bias_multiclass_plots import accuracy_bar_plot, frequency_matrix_plot, frequency_plot
from IPython.display import display


def binary_classification_data_exploration(dataset):
    from io import StringIO

    from IPython.core.display import HTML
    from ipywidgets import embed  # type: ignore

    output = widgets.Output()

    def update_plot(change):
        with output:
            output.clear_output(wait=True)
            fig, axs = plt.subplots(1, 2, figsize=(15, 5))

            group_pie_plot(dataset["p_attr"][change["new"]], ax=axs[0])
            axs[0].set_title(f"Protected Attribute/Group: {change['new']}")
            axs[0].grid(True)

            frequency_plot(dataset["p_attr"][change["new"]], dataset["y"], ax=axs[1])
            axs[1].set_title(f"Positive Rate by Group: {change['new']}")
            axs[1].grid(True)

            plt.show()

    protected_attributes = dataset["p_attr"].columns
    dropdown = widgets.Dropdown(options=protected_attributes, value=protected_attributes[0], description="p_attr:")
    dropdown.observe(update_plot, names="value")
    content = widgets.VBox([dropdown, output])
    display(content)
    update_plot({"new": dropdown.value})

    html_stream = StringIO()
    embed.embed_minimal_html(html_stream, views=[content], title="Mi widget")

    # Obtener el contenido HTML como una cadena
    html_content = html_stream.getvalue()

    # Mostrar el HTML
    display(HTML(html_content))


def clustering_data_exploration(dataset):
    output = widgets.Output()

    def update_plot(change):
        with output:
            output.clear_output(wait=True)
            fig, axs = plt.subplots(1, 2, figsize=(15, 5))

            group_pie_plot(dataset["p_attr"][change["new"]], ax=axs[0])
            axs[0].set_title(f"Protected Attribute/Group: {change['new']}")
            axs[0].grid(True)

            histogram_plot(dataset["p_attr"][change["new"]], dataset["y"], ax=axs[1])
            axs[1].set_title(f"Clusters by Group: {change['new']}")
            axs[1].grid(True)

            plt.show()

    protected_attributes = dataset["p_attr"].columns
    dropdown = widgets.Dropdown(options=protected_attributes, value=protected_attributes[0], description="p_attr:")
    dropdown.observe(update_plot, names="value")
    content = widgets.VBox([dropdown, output])
    display(content)
    update_plot({"new": dropdown.value})


def binary_classification_model_exploration(dataset, y_pred):
    output = widgets.Output()

    def update_plot(change):
        with output:
            output.clear_output(wait=True)
            fig, axs = plt.subplots(1, 2, figsize=(15, 5))

            # group_pie_plot(dataset['p_attr'][change['new']], ax=axs[0])
            accuracy_bar_plot(dataset["p_attr"][change["new"]], y_pred, dataset["y"], ax=axs[0])
            axs[0].set_title(f"Accuracy Bar Plot: {change['new']}")
            axs[0].grid(True)

            frequency_plot(dataset["p_attr"][change["new"]], y_pred, ax=axs[1])
            axs[1].set_title(f"Positive Rate by Group: {change['new']}")
            axs[1].grid(True)

            plt.show()

    protected_attributes = dataset["p_attr"].columns
    dropdown = widgets.Dropdown(options=protected_attributes, value=protected_attributes[0], description="p_attr:")
    dropdown.observe(update_plot, names="value")
    content = widgets.VBox([dropdown, output])
    display(content)
    update_plot({"new": dropdown.value})


def clustering_model_exploration(dataset, y_pred):
    output = widgets.Output()

    def update_plot(change):
        with output:
            output.clear_output(wait=True)
            fig, ax = plt.subplots(1, 1, figsize=(10, 10))

            frequency_matrix_plot(dataset["p_attr"][change["new"]], y_pred, normalize="group", ax=ax)
            ax.set_title(f"Group-Clusters Frequency: {change['new']}")
            ax.grid(True)

            plt.show()

    protected_attributes = dataset["p_attr"].columns
    dropdown = widgets.Dropdown(options=protected_attributes, value=protected_attributes[0], description="p_attr:")
    dropdown.observe(update_plot, names="value")
    content = widgets.VBox([dropdown, output])
    display(content)
    update_plot({"new": dropdown.value})


def bias_data_exploration(learning_task, dataset):
    if learning_task == "binary_classification":
        return binary_classification_data_exploration(dataset)
    if learning_task == "clustering":
        return clustering_data_exploration(dataset)
    return None


def bias_model_exploration(learning_task, dataset, y_pred):
    if learning_task == "binary_classification":
        return binary_classification_model_exploration(dataset, y_pred)
    if learning_task == "clustering":
        return clustering_model_exploration(dataset, y_pred)
    return None



from __future__ import annotations

import numpy as np
from holisticai.utils import ConditionalImportances, Importances


class RankAlignment:
    name: str = "Rank Alignment"
    reference: float = 1.0

    def __call__(self, conditional_feature_importance: ConditionalImportances, feature_importance: Importances):
        """
        Calculates the rank alignment metric between conditional feature importance and overall feature importance.

        Args:
            conditional_feature_importance (ConditionalImportances): Conditional feature importance object.
            feature_importance (Importances): Overall feature importance object.

        Returns:
            float: The rank alignment metric value.
        """
        feature_names = feature_importance.feature_names
        conditional_position_parity = {}
        for group_name, cond_features in conditional_feature_importance:
            cond_feature_names = cond_features.feature_names
            intersections = []
            for top_k in range(1, len(feature_importance) + 1):
                ggg = set(cond_feature_names[:top_k])
                vvv = set(feature_names[:top_k])
                u = len(set(ggg).intersection(vvv)) / top_k
                intersections.append(u)
            conditional_position_parity[group_name] = np.mean(intersections)
        return float(np.mean(np.mean(list(conditional_position_parity.values()))))


def rank_alignment(conditional_feature_importance: ConditionalImportances, ranked_feature_importance: Importances):
    """
    The `Rank Alignment` metric is used to measure the alignment between conditional feature importance and overall feature importance.
    It provides a way to assess how well the importance rankings of features in a conditional context match the rankings in the overall context.

    Parameters
    ----------
    conditional_feature_importance: ConditionalImportances
        The feature importance for each output label (classification) or output region (regression).
    ranked_feature_importance: Importances
        The ranked feature importance values.

    Returns
    -------
        float: The rank alignment metric value.

    Example
    -------
    >>> from holisticai.explainability.commons import (
    ...     ConditionalImportances,
    ...     Importances,
    ... )
    >>> from holisticai.explainability.metrics import rank_alignment
    >>> conditional_feature_importance = ConditionalImportances(
    ...     values={
    ...         "0": Importances(
    ...             values=[0.1, 0.2, 0.3, 0.4],
    ...             feature_names=["feature_2", "feature_3", "feature_4"],
    ...         ),
    ...         "1": Importances(
    ...             values=[0.4, 0.3, 0.2, 0.1],
    ...             feature_names=["feature_1", "feature_2", "feature_3", "feature_4"],
    ...         ),
    ...     }
    ... )
    >>> ranked_feature_importance = Importances(
    ...     values=[0.5, 0.3, 0.2],
    ...     feature_names=["feature_1", "feature_2", "feature_3"],
    ... )
    >>> rank_alignment(conditional_feature_importance, ranked_feature_importance)
    0.6944444444444444
    """
    metric = RankAlignment()
    return metric(conditional_feature_importance, ranked_feature_importance)



# Imports
import numpy as np
from numpy.testing import assert_approx_equal

# Clustering
from holisticai.bias.metrics import (
    cluster_balance,
    cluster_dist_entropy,
    cluster_dist_kl,
    cluster_dist_l1,
    min_cluster_ratio,
    silhouette_diff,
    social_fairness_ratio,
)

# Format data into numpy
group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
y_pred_cl = np.array([0, 1, 1, 2, 0, 0, 0, 0, 1, 2])
y_true_cl = np.array([0, 1, 0, 2, 0, 0, 1, 0, 1, 2])
X = np.array(
    [
        [-1, 1],
        [1, 1],
        [1, 1],
        [0, -1],
        [-1, 1],
        [-1, 1],
        [-1, 1],
        [-1, 1],
        [1, 1],
        [0, -1],
    ]
)
centroids = np.array([[-2, 1], [1, 2], [0, -2]])

def test_cluster_balance():
    """test cluster_balance"""
    a = cluster_balance(group_a, group_b, y_pred_cl)
    b = 0.5
    assert_approx_equal(a, b)


def test_min_cluster_ratio():
    """test min_cluster_ratio"""
    a = min_cluster_ratio(group_a, group_b, y_pred_cl)
    b = 1 / 4
    assert_approx_equal(a, b)


def test_avg_cluster_ratio():
    """test avg_cluster_ratio
    a = avg_cluster_ratio(group_a, group_b, y_pred_cl)
    b = 1 / 12 + 1
    """
    assert True


def test_cluster_dist_l1():
    """test cluster_dist_l1"""
    a = cluster_dist_l1(group_a, group_b, y_pred_cl)
    b = 0.5 * (abs(1 / 4 - 2 / 3) + abs(1 / 2 - 1 / 6) + abs(1 / 4 - 1 / 6))
    assert_approx_equal(a, b)


def test_cluster_dist_kl():
    """test cluster_dist_kl"""
    a = cluster_dist_kl(group_a, group_b, y_pred_cl)
    b = 0.40546510810816444
    assert_approx_equal(a, b)


def test_cluster_dist_entropy():
    """test cluster_dist_entropy"""
    a = cluster_dist_entropy(group_a, y_pred_cl)
    b = 1.5 * np.log(2)
    assert_approx_equal(a, b)


def test_social_fairness_ratio():
    """test social_fairness_ratio"""
    a = social_fairness_ratio(group_a, group_b, X, centroids)
    b = 1
    assert_approx_equal(a, b)


def test_silhouette_diff():
    """test silhouette_diff"""
    a = silhouette_diff(group_a, group_b, X, y_pred_cl)
    b = 0
    assert_approx_equal(a, b)



from holisticai.utils.feature_importances._conditional import group_samples_by_learning_task
from holisticai.utils.feature_importances._lime import LIMEImportanceCalculator, compute_lime_feature_importance
from holisticai.utils.feature_importances._permutation_feature_importance import (
    PermutationFeatureImportanceCalculator,
    compute_permutation_feature_importance,
)
from holisticai.utils.feature_importances._shap import SHAPImportanceCalculator, compute_shap_feature_importance
from holisticai.utils.feature_importances._surrogate_feature_importance import (
    SurrogateFeatureImportanceCalculator,
    compute_surrogate_feature_importance,
)

__all__ = [
    "PermutationFeatureImportanceCalculator",
    "compute_permutation_feature_importance",
    "group_samples_by_learning_task",
    "SurrogateFeatureImportanceCalculator",
    "compute_surrogate_feature_importance",
    "LIMEImportanceCalculator",
    "compute_lime_feature_importance",
    "SHAPImportanceCalculator",
    "compute_shap_feature_importance",
]



import jax.numpy as jnp
from jax import vmap


class KNeighborsClassifier:
    def fit(self, X_train, y_train):
        self.X_train = jnp.array(X_train)
        self.y_train = jnp.array(y_train)
        self.n_train_samples = len(X_train)

    def kneighbors_batched(self, X, n_neighbors=None, return_distance=True, batch_size=100):
        if n_neighbors is None:
            n_neighbors = self.n_train_samples

        n_samples = X.shape[0]

        for i in range(0, n_samples, batch_size):
            X_batch = jnp.array(X[i : i + batch_size])
            dists_batch = jnp.sqrt(jnp.sum((X_batch[:, None, :] - self.X_train[None, :, :]) ** 2, axis=-1))
            batch_neighbors_idx = jnp.argsort(dists_batch, axis=1)[:, :n_neighbors]
            if return_distance:
                yield dists_batch, batch_neighbors_idx
            else:
                yield batch_neighbors_idx

    def kneighbors(self, X, n_neighbors=None, return_distance=True, batch_size=100):
        neighbors_dists = []
        neighbors_idx = []

        for dists_batch, batch_neighbors_idx in self.kneighbors_batched(X, n_neighbors, return_distance, batch_size):
            batch_neighbors_dists = jnp.take_along_axis(dists_batch, batch_neighbors_idx, axis=1)
            neighbors_dists.append(batch_neighbors_dists)
            neighbors_idx.append(batch_neighbors_idx)

        neighbors_idx = jnp.vstack(neighbors_idx)
        if return_distance:
            neighbors_dists = jnp.vstack(neighbors_dists)
            return neighbors_dists, neighbors_idx
        return neighbors_idx

    def _predict_single(self, x):
        # Compute distances from x to all training points
        dists = jnp.sqrt(jnp.sum((self.X_train - x) ** 2, axis=-1))
        # Find the k nearest neighbors
        neighbors = jnp.argsort(dists)[: self.n_neighbors]
        # Predict the label by majority vote
        unique, counts = jnp.unique(self.y_train[neighbors], return_counts=True)
        return unique[jnp.argmax(counts)]

    def predict(self, X):
        # Vectorize prediction for all input points
        predictions = vmap(self._predict_single)(jnp.array(X))
        return predictions



from __future__ import annotations

import numpy as np
import pandas as pd
from holisticai.utils import Importances, PartialDependence


def compute_feature_scores(data, threshold):
    scores = [
        {
            "feature": feat,
            "scores": sum([1 if rr > threshold else 0 for rr in r]),
            "few_points": flag,
        }
        for feat, (r, flag) in data.items()
    ]
    scores = pd.DataFrame(scores)[["few_points", "feature", "scores"]]
    return scores.sort_values("scores", ascending=False)


def calculate_discrete_derivative(y_values):
    """Calculate the discrete derivative for a sequence of y values."""
    dy = np.diff(y_values)
    dx = np.ones_like(dy)  # Assuming x values are equally spaced with a difference of 1
    return dy / dx


def cosine_similarity(v1, v2):
    """Calculate the cosine similarity between two vectors."""
    dot_product = np.dot(v1, v2)
    norm_v1 = np.linalg.norm(v1)
    norm_v2 = np.linalg.norm(v2)
    return dot_product / (norm_v1 * norm_v2)


def compare_tangents(points):
    num_sections = 3
    few_points = False
    n = len(points)
    tol = 1e-5
    if n < num_sections:
        few_points = True
        return (1, 1), few_points

    cut1 = n // 3
    cut2 = 2 * n // 3

    section1 = points[: cut1 + 1]
    section2 = points[cut1 : cut2 + 1]
    section3 = points[cut2:]

    sections = [section1, section2, section3]
    slopes = []
    for section in sections:
        if len(section) > 1:
            section_slopes = calculate_discrete_derivative(section)
            average_slope = np.mean(section_slopes)
        else:
            average_slope = 0
        slopes.append(average_slope + tol)

    similarities = []
    for i in range(len(slopes) - 1):
        similarity = cosine_similarity([slopes[i]], [slopes[i + 1]])
        similarities.append(similarity)
    return similarities, few_points


class XAIEaseAnnotator:
    threshold: float = 0
    levels: list[str] = ["Hard", "Medium", "Easy"]

    def compute_xai_ease_score_data(self, partial_dependence, ranked_feature_importance):
        """
        Computes the XAI Ease Score data for a given partial dependence plot.

        Args:
            partial_dependence (dict): A dictionary containing the partial dependence plots for each feature.

        Returns:
            score_data (DataFrame): The computed score data.
        """
        partial_dependence_formatted = {
            f: partial_dependence[i]["average"][0] for i, f in enumerate(ranked_feature_importance.feature_names)
        }
        data = {feat: compare_tangents(df) for feat, df in partial_dependence_formatted.items()}
        score_data = compute_feature_scores(data, self.threshold)
        score_data["scores"] = score_data.apply(lambda x: self.levels[x["scores"]], axis=1)
        return score_data


class XAIEaseScore:
    """
    Class for computing the XAI Ease Score.

    The XAI Ease Score measures the ease of interpretability of a model's explanations.
    It takes into account the similarity between partial dependence plots of different features
    and assigns scores based on the similarity values.

    Attributes:
        num_chunks (int): The number of chunks to divide the partial dependence plots into.
        threshold (float): The threshold value for computing feature scores.
        levels (list): The levels of ease scores, in descending order of difficulty.

    Methods:
        __compute_xai_ease_score: Computes the XAI Ease Score for a given score data.
        __call__: Computes the XAI Ease Score for a set of partial dependence plots.
        __xai_feature_ease_score: Computes the XAI Ease Score for a single partial dependence plot.
    """

    reference: float = 1.0
    name: str = "XAI Ease Score"
    annotator: XAIEaseAnnotator = XAIEaseAnnotator()

    def __init__(self, detailed: bool = False):
        self.detailed = detailed

    def compute_xai_ease_score(self, score_data):
        """
        Computes the XAI Ease Score for a given score data.

        Args:
            score_data (DataFrame): The score data.

        Returns:
            xai_ease_score (float): The computed XAI Ease Score.
        """
        max_score = 2
        score_dict = pd.DataFrame(
            score_data.groupby("scores").count()["feature"] / score_data.groupby("scores").count()["feature"].sum()
        ).to_dict()["feature"]

        values = []
        full_score = {c: 0 for c in self.annotator.levels}
        for c, v in score_dict.items():
            full_score[c] = v
            values.append(self.annotator.levels.index(c) * full_score[c])

        return sum(values) / max_score

    def __call__(
        self,
        partial_dependence: PartialDependence,
        ranked_feature_importance: Importances,
    ):
        """
        Computes the XAI Ease Score for a set of partial dependence plots.

        Args:
            partial_dependence (list): A list of dictionaries containing the partial dependence plots for each feature.
            features (list): A list of feature names.

        Returns:
            float: The computed XAI Ease Score.
        """

        def compute_metric(pdep, rfi):
            score_data = self.annotator.compute_xai_ease_score_data(pdep, rfi)
            return float(self.compute_xai_ease_score(score_data))

        scores = [compute_metric(pdep, ranked_feature_importance) for pdep in partial_dependence.values]
        if self.detailed:
            return scores
        return float(np.mean(scores))


def xai_ease_score(partial_dependence: PartialDependence, ranked_feature_importance: Importances):
    """
    This metric, ranging from 0 to 1, measures the ease of explaining a model's predictions for the top feature importances (>80%) \
    using partial dependence plots. The XAI Ease Score considers the similarity between regions of the partial dependence plots \
    for different features and assigns scores based on these similarity values.

    Parameters
    ----------
    partial_dependence: PartialDependence
        The partial dependence values for each feature.
    ranked_feature_importance: Importances
        The ranked feature importance values.

    Returns
    -------
        float: The XAI Ease Score.

    Examples
    --------
    >>> from holisticai.explainability.commons import PartialDependence, Importances
    >>> from holisticai.explainability.metrics.global_importance import xai_ease_score
    >>> partial_dependence = [
    ...     {
    ...         "average": [[0.1, 0.2, 0.3, 0.1, 0.2, 0.3, 0.1, 0.2, 0.3]],
    ...         "grid_values": [[1, 2, 3, 4, 5, 6, 7, 8, 9]],
    ...     },
    ...     {
    ...         "average": [[0.4, 0.5, 0.6, 0.6, 0.5, 0.4, 0.4, 0.5, 0.6]],
    ...         "grid_values": [[1, 2, 3, 4, 5, 6, 7, 8, 9]],
    ...     },
    ... ]
    >>> partial_dependence = PartialDependence(values=partial_dependence)
    >>> feature_importance = Importances(values=np.array([0.5, 0.5]),
    >>> feature_names=['feature1', 'feature2'])
    >>> xai_ease_score(partial_dependence, feature_importance)
    0.5
    """
    metric = XAIEaseScore()
    return metric(partial_dependence, ranked_feature_importance)



import numpy as np
from holisticai.utils import Importances
from scipy.spatial.distance import jensenshannon
from scipy.stats import entropy


class ImportanceSpread:
    def __call__(self, feature_importance):
        """
        Parameters
        ----------
        feature_importance: np.array
            array with raw feature importance
        divergence: bool
            if True calculate the inverse Jensen-Shannon divergence, otherwise the ratio
        """
        tol = 1e-8
        feature_importances = np.array(feature_importance.values, dtype=float)
        if len(feature_importances) == 0 or sum(feature_importances) < tol:
            return 0 if self.divergence else 1

        importance = feature_importances
        feature_weight = importance / sum(importance)
        feature_equal_weight = np.array([1.0 / len(importance)] * len(importance))

        if self.divergence:
            metric = 1 - jensenshannon(feature_weight, feature_equal_weight, base=2)
        else:
            metric = entropy(feature_weight) / entropy(feature_equal_weight)
        return float(metric)


class SpreadRatio(ImportanceSpread):
    name: str = "Spread Ratio"
    reference: float = 0
    divergence: bool = False


class SpreadDivergence(ImportanceSpread):
    name: str = "Spread Divergence"
    reference: float = 0
    divergence: bool = True


def spread_ratio(feature_importance: Importances):
    """
    The spread ratio, ranging from 0 to 1, measures the degree of evenness or concentration in the distribution of feature importance values.
    A higher spread ratio indicates a more evenly distributed feature importance, while a lower spread ratio indicates a more concentrated feature importance.
    A lower ratio concentrates the importances and facilitates interpretability.

    Parameters
    ----------
    feature_importance: Importances
        The feature importance values for the features.

    Returns
    -------
    float:
        The spread ratio of the feature importance.

    Examples
    --------

    >>> from holisticai.explainability.commons import Importances
    >>> from holisticai.explainability.metrics.global_importance import spread_ratio
    >>> values = np.array([0.10, 0.20, 0.30])
    >>> feature_names = ["feature_1", "feature_2", "feature_3"]
    >>> feature_importance = Importances(values=values, feature_names=feature_names)
    >>> score = spread_ratio(feature_importance)
    0.9206198357143052

    """
    metric = SpreadRatio()
    return metric(feature_importance)


def spread_divergence(feature_importance: Importances):
    """
    Calculates the spread divergence metric based on the inverse of the Jensen-Shannon distance
    (square root of the Jensen-Shannon divergence), for a given feature importance.
    A lower ratio concentrates the importances and facilitates interpretability.

    Parameters
    ----------
    feature_importance: Importances
        The feature importance values for the features.

    Returns
    -------
        float: The spread divergence metric value.

    Example
    -------

    >>> from holisticai.explainability.commons import Importances
    >>> from holisticai.explainability.metrics.global_importance import (
    ...     spread_divergence,
    ... )
    >>> feature_importance = Importances(
    ...     values=np.array([0.10, 0.20, 0.30]),
    ...     feature_names=["feature_1", "feature_2", "feature_3"],
    ... )
    >>> score = spread_divergence(feature_importance)
    0.8196393599933761

    """
    metric = SpreadDivergence()
    return metric(feature_importance)
